import itertools
from pathlib import Path
from tempfile import TemporaryDirectory
from urllib.error import HTTPError, URLError

import nibabel as nib
import numpy as np
import numpy.testing as npt
import pytest

from dipy.data import get_fnames
from dipy.io.surface import load_surface, save_surface
from dipy.io.utils import Origin, Space
from dipy.utils.optpkg import optional_package

vtk, have_vtk, setup_module = optional_package(
    "vtk", min_version="9.0.0", max_version="9.1.0"
)

FOLDERS_GII = ["ascii", "base64", "gzip_base64"]
FILENAMES_GII = ["pial.L.surf.gii", "smoothwm.L.surf.gii"]

FOLDERS = ["big_affine_freesurfer", "small_affine_freesurfer"]
HEMISPHERES = ["lh", "rh"]
TYPES = ["pial", "smoothwm", "orig"]

SPACES = [Space.RASMM, Space.LPSMM, Space.VOXMM, Space.VOX]
ORIGINS = [Origin.NIFTI, Origin.TRACKVIS]


def setup_module():
    global FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA
    try:
        FILEPATH_DIX = get_fnames(name="real_data_io")
    except (HTTPError, URLError) as e:
        FILEPATH_DIX, POINTS_DATA, STREAMLINES_DATA = None, None, None
        error_msg = f'"Tests Data failed to download." Reason: {e}'
        pytest.skip(error_msg, allow_module_level=True)
        return


@pytest.mark.skipif(not have_vtk, reason="Requires VTK")
def test_pial_load_save():
    data_raw = nib.freesurfer.read_geometry(FILEPATH_DIX["naf_lh.pial"])

    sfs = load_surface(
        FILEPATH_DIX["naf_lh.pial"], FILEPATH_DIX["naf_mni_masked.nii.gz"]
    )

    # Change the space/origin, should not affect the saved pial
    sfs.to_vox()
    sfs.to_corner()

    with TemporaryDirectory() as tmpdir:
        save_surface(
            sfs, Path(tmpdir) / "lh.pial", ref_pial=FILEPATH_DIX["naf_lh.pial"]
        )
        data_save = nib.freesurfer.read_geometry(Path(tmpdir) / "lh.pial")
    npt.assert_almost_equal(data_raw[0], data_save[0], decimal=5)


@pytest.mark.skipif(not have_vtk, reason="Requires VTK")
@pytest.mark.parametrize("space,origin", list(itertools.product(SPACES, ORIGINS)))
def test_vtk_matching_space(space, origin):
    sfs = load_surface(
        FILEPATH_DIX["naf_lh.pial"], FILEPATH_DIX["naf_mni_masked.nii.gz"]
    )
    sfs.to_rasmm()
    sfs.to_center()
    ref_vertices = sfs.vertices.copy()

    with TemporaryDirectory() as tmpdir:
        save_surface(sfs, Path(tmpdir) / "tmp.vtk", to_space=space, to_origin=origin)
        sfs = load_surface(
            Path(tmpdir) / "tmp.vtk",
            FILEPATH_DIX["naf_mni_masked.nii.gz"],
            from_space=space,
            from_origin=origin,
        )

        sfs.to_rasmm()
        sfs.to_center()
        save_vertices = sfs.vertices.copy()
        npt.assert_almost_equal(ref_vertices, save_vertices, decimal=5)


@pytest.mark.skipif(not have_vtk, reason="Requires VTK")
@pytest.mark.parametrize(
    "type,fname,space,origin",
    list(itertools.product(FOLDERS_GII, FILENAMES_GII, SPACES, ORIGINS)),
)
def test_gifti_matching_space(type, fname, space, origin):
    if type == "gzip_base64":
        fname += ".gz"
    sfs = load_surface(FILEPATH_DIX[fname], FILEPATH_DIX["anat.nii.gz"])
    sfs.to_rasmm()
    sfs.to_center()
    ref_vertices = sfs.vertices.copy()

    with TemporaryDirectory() as tmpdir:
        save_surface(sfs, Path(tmpdir) / "tmp.gii", to_space=space, to_origin=origin)
        sfs = load_surface(
            Path(tmpdir) / "tmp.gii",
            FILEPATH_DIX["anat.nii.gz"],
            from_space=space,
            from_origin=origin,
        )

        sfs.to_rasmm()
        sfs.to_center()
        save_vertices = sfs.vertices.copy()
        npt.assert_almost_equal(ref_vertices, save_vertices, decimal=5)


@pytest.mark.skipif(not have_vtk, reason="Requires VTK")
@pytest.mark.parametrize(
    "dataset,hemisphere,type", list(itertools.product(FOLDERS, HEMISPHERES, TYPES))
)
def test_freesurfer_density_operation(dataset, hemisphere, type):
    prefix = "baf" if dataset == "big_affine_freesurfer" else "saf"
    fname = f"{prefix}_{hemisphere}.{type}"
    sfs = load_surface(FILEPATH_DIX[fname], FILEPATH_DIX[f"{prefix}_t1.nii.gz"])

    assert sfs.is_bbox_in_vox_valid()

    data = np.zeros(sfs.dimensions, dtype=np.uint32)

    sfs.to_vox()
    sfs.to_corner()

    # Compute density map in the numpy grid
    for vertice in sfs.vertices:
        coord = tuple(vertice.astype(np.int32))
        data[coord] += 1

    with TemporaryDirectory() as tmpdir:
        nib.save(
            nib.Nifti1Image(data, sfs.affine),
            Path(tmpdir) / f"{hemisphere}_{type}.nii.gz",
        )

    # Compute the barycenter of the density map and compare it to the
    # approximate barycenter
    barycenter = np.mean(np.argwhere(data), axis=0)
    if dataset == "small_affine_freesurfer":
        approx_barycenter = [141, 100, 82] if hemisphere == "lh" else [80, 101, 83]
    elif dataset == "big_affine_freesurfer":
        approx_barycenter = [139, 96, 80] if hemisphere == "lh" else [79, 97, 78]

    npt.assert_(np.linalg.norm(barycenter - approx_barycenter) < 2.0)



# cython: boundscheck=False
# cython: cdivision=True
# cython: initializedcheck=False
# cython: wraparound=False
"""
================================
Non-Local Means Denoising
================================

"""
import numpy as np
cimport numpy as cnp

cimport cython

from cython.parallel import parallel, prange, threadid
from dipy.utils.omp import determine_num_threads
from dipy.utils.omp cimport set_num_threads, restore_default_num_threads

from libc.math cimport sqrt, exp
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy



def nlmeans_3d_classic(arr, mask=None, sigma=None, patch_radius=1,
               block_radius=5, rician=True, num_threads=None):
    """ Non-local means for denoising 3D images

    Parameters
    ----------
    arr : 3D ndarray
        The array to be denoised
    mask : 3D ndarray
    sigma : float or 3D array
        standard deviation of the noise estimated from the data
    patch_radius : int
        patch size is ``2 x patch_radius + 1``. Default is 1.
    block_radius : int
        block size is ``2 x block_radius + 1``. Default is 5.
    rician : boolean
        If True the noise is estimated as Rician, otherwise Gaussian noise
        is assumed.
    num_threads : int, optional
        Number of threads to be used for OpenMP parallelization. If None
        (default) the value of OMP_NUM_THREADS environment variable is used
        if it is set, otherwise all available threads are used. If < 0 the
        maximal number of threads minus $|num_threads + 1|$ is used (enter -1 to
        use as many threads as possible). 0 raises an error.

    Returns
    -------
    denoised_arr : ndarray
        the denoised ``arr`` which has the same shape as ``arr``.
    """

    if arr.ndim != 3:
        raise ValueError(f'data needs to be a 3D ndarray, got shape {arr.shape}')

    if mask is None:
        mask = np.ones(arr.shape, dtype='f8')
    else:
        mask = np.ascontiguousarray(mask, dtype='f8')

    if mask.ndim != 3:
        raise ValueError(f'mask needs to be a 3D ndarray, got shape {mask.shape}')

    if sigma.ndim != 3:
        raise ValueError(f'sigma needs to be a 3D ndarray, got shape {sigma.shape}')

    arr = np.ascontiguousarray(arr, dtype='f8')
    arr = add_padding_reflection(arr, block_radius)
    mask = add_padding_reflection(mask.astype('f8'), block_radius)
    sigma = np.ascontiguousarray(sigma, dtype='f8')
    sigma = add_padding_reflection(sigma.astype('f8'), block_radius)
    arrnlm = _nlmeans_3d(arr, mask, sigma, patch_radius, block_radius,
                         rician, num_threads)

    return remove_padding(arrnlm, block_radius)


def _nlmeans_3d(double[:, :, ::1] arr, double[:, :, ::1] mask,
                double[:, :, ::1] sigma, patch_radius=1, block_radius=5,
                rician=True, num_threads=None):
    """ This algorithm denoises the value of every voxel (i, j, k) by
    calculating a weight between a moving 3D patch and a static 3D patch
    centered at (i, j, k). The moving patch can only move inside a
    3D block.
    """

    cdef:
        cnp.npy_intp i, j, k, I, J, K
        double[:, :, ::1] out = np.zeros_like(arr)
        double summ = 0
        cnp.npy_intp P = patch_radius
        cnp.npy_intp B = block_radius
        int threads_to_use = -1

    I = arr.shape[0]
    J = arr.shape[1]
    K = arr.shape[2]

    threads_to_use = determine_num_threads(num_threads)
    set_num_threads(threads_to_use)

    # move the block
    with nogil, parallel():

        for i in prange(B, I - B, schedule="dynamic"):
            for j in range(B, J - B):
                for k in range(B, K - B):

                    if mask[i, j, k] == 0:
                        continue

                    out[i, j, k] = process_block(arr, i, j, k, B, P, sigma)

    if num_threads is not None:
        restore_default_num_threads()

    new = np.asarray(out)

    if rician:
        new -= 2 * np.asarray(sigma)**2
        new[new < 0] = 0

    return np.sqrt(new)


cdef double process_block(double[:, :, ::1] arr,
                          cnp.npy_intp i, cnp.npy_intp j, cnp.npy_intp k,
                          cnp.npy_intp B, cnp.npy_intp P, double[:, :, ::1] sigma) nogil:
    """ Process the block with center at (i, j, k)

    Parameters
    ----------
    arr : 3D array
        C contiguous array of doubles
    i, j, k : int
        center of block
    B : int
        block radius
    P : int
        patch radius
    sigma : 3D array
        local noise standard deviation

    Returns
    -------
    new_value : double
    """

    cdef:
        cnp.npy_intp m, n, o, M, N, O, patch_vol_size, a, b, c, cnt, step
        double summ, d, w, sumw, sum_out, x, sigm
        double * W
        double * cache
        double * sigma_block
        double denom
        cnp.npy_intp BS = B * 2 + 1

    cnt = 0
    sumw = 0
    patch_vol_size = (P + P + 1) * (P + P + 1) * (P + P + 1)

    W = <double *> malloc(BS * BS * BS * sizeof(double))
    cache = <double *> malloc(BS * BS * BS * sizeof(double))
    sigma_block = <double *> malloc(BS * BS * BS * sizeof(double))

    # (i, j, k) coordinates are the center of the static patch
    # copy block in cache
    copy_block_3d(cache, BS, BS, BS, arr, i - B, j - B, k - B)
    copy_block_3d(sigma_block, BS, BS, BS, sigma, i - B, j - B, k - B)

    # calculate weights between the central patch and the moving patch in block
    # (m, n, o) coordinates are the center of the moving patch
    # (a, b, c) run inside both patches
    for m in range(P, BS - P):
        for n in range(P, BS - P):
            for o in range(P, BS - P):

                summ = 0
                sigm = 0

                # calculate square distance
                for a in range(-P, P + 1):
                    for b in range(-P, P + 1):
                        for c in range(-P, P + 1):

                            # this line takes most of the time! mem access
                            d = cache[(B + a) * BS * BS + (B + b) * BS + (B + c)] - \
                                cache[(m + a) * BS * BS + (n + b) * BS + (o + c)]
                            summ += d * d
                            sigm += sigma_block[(m + a) *
                                                BS * BS + (n + b) * BS + (o + c)]

                denom = sqrt(2) * (sigm / patch_vol_size)**2
                w = exp(-(summ / patch_vol_size) / denom)
                sumw += w
                W[cnt] = w
                cnt += 1

    cnt = 0
    sum_out = 0

    # calculate normalized weights and sums of the weights with the positions
    # of the patches
    for m in range(P, BS - P):
        for n in range(P, BS - P):
            for o in range(P, BS - P):

                if sumw > 0:
                    w = W[cnt] / sumw
                else:
                    w = 0

                x = cache[m * BS * BS + n * BS + o]
                sum_out += w * x * x
                cnt += 1

    free(W)
    free(cache)
    free(sigma_block)

    return sum_out


def add_padding_reflection(double[:, :, ::1] arr, padding):
    cdef:
        double[:, :, ::1] final
        cnp.npy_intp i, j, k
        cnp.npy_intp B = padding
        cnp.npy_intp[::1] indices_i = correspond_indices(arr.shape[0], padding)
        cnp.npy_intp[::1] indices_j = correspond_indices(arr.shape[1], padding)
        cnp.npy_intp[::1] indices_k = correspond_indices(arr.shape[2], padding)

    final = np.zeros(
        np.array(
            (arr.shape[0],
             arr.shape[1],
             arr.shape[2])) +
        2 *
        padding)

    for i in range(final.shape[0]):
        for j in range(final.shape[1]):
            for k in range(final.shape[2]):
                final[i, j, k] = arr[indices_i[i], indices_j[j], indices_k[k]]

    return final


def correspond_indices(dim_size, padding):
    return np.ascontiguousarray(np.hstack((np.arange(1,
                                                     padding + 1)[::-1],
                                           np.arange(dim_size),
                                           np.arange(dim_size - padding - 1,
                                                     dim_size - 1)[::-1])),
                                dtype=np.intp)


def remove_padding(arr, padding):
    shape = arr.shape
    return arr[padding:shape[0] - padding,
               padding:shape[1] - padding,
               padding:shape[2] - padding]


@cython.wraparound(False)
@cython.boundscheck(False)
cdef cnp.npy_intp copy_block_3d(double * dest,
                                cnp.npy_intp I,
                                cnp.npy_intp J,
                                cnp.npy_intp K,
                                double[:, :, ::1] source,
                                cnp.npy_intp min_i,
                                cnp.npy_intp min_j,
                                cnp.npy_intp min_k) nogil:

    cdef cnp.npy_intp i, j

    for i in range(I):
        for j in range(J):
            memcpy(&dest[i * J * K  + j * K], &source[i + min_i, j + min_j, min_k], K * sizeof(double))

    return 1


# ==== BLOCKWISE NON-LOCAL MEANS IMPLEMENTATION ====

cdef inline int _int_max(int a, int b):
    return a if a >= b else b
cdef inline int _int_min(int a, int b):
    return a if a <= b else b


def _firdn_vector(double[:] f, double[:] h, double[:] out):
    cdef int n = len(f)
    cdef int klen = len(h)
    cdef int outLen = (n + klen) // 2
    cdef double ss
    cdef int i, k, limInf, limSup, x = 0, ox = 0, ks = 0
    for i in range(outLen):
        ss = 0
        limInf = _int_max(0, x - klen + 1)
        limSup = 1 + _int_min(n - 1, x)
        ks = limInf
        for k in range(limInf, limSup):
            ss += f[ks] * h[x - k]
            ks += 1
        out[ox] = ss
        x += 2
        ox += 1


def _upfir_vector(double[:] f, double[:] h, double[:] out):
    cdef int n = f.shape[0]
    cdef int klen = h.shape[0]
    cdef int outLen = 2 * n + klen - 2
    cdef int x, limInf, limSup, k, ks
    cdef double ss
    for x in range(outLen):
        limInf = _int_max(0, x - klen + 1)
        if limInf % 2 == 1:
            limInf += 1
        limSup = _int_min(2 * (n - 1), x)
        if limSup % 2 == 1:
            limSup -= 1
        ss = 0
        k = limInf
        ks = limInf // 2
        while k <= limSup:
            ss += f[ks] * h[x - k]
            k += 2
            ks += 1
        out[x] = ss


def _firdn_matrix(double[:, :] F, double[:] h, double[:, :] out):
    cdef int n = F.shape[0]
    cdef int m = F.shape[1]
    cdef int j
    for j in range(m):
        _firdn_vector(F[:, j], h, out[:, j])


def _upfir_matrix(double[:, :] F, double[:] h, double[:, :] out):
    cdef int n = F.shape[0]
    cdef int m = F.shape[1]
    for j in range(m):
        _upfir_vector(F[:, j], h, out[:, j])


cpdef firdn(double[:, :] image, double[:] h):
    """
    Applies the filter given by the convolution kernel 'h' columnwise to
    'image', then subsamples by 2. This is a special case of the matlab's
    'upfirdn' function, ported to python. Returns the filtered image.

    Parameters
    ----------
    image: 2D array of doubles
        the input image to be filtered
    h: double array
        the convolution kernel
    """
    nrows = image.shape[0]
    ncols = image.shape[1]
    ll = h.shape[0]
    cdef double[:, :] filtered = np.zeros(shape=((nrows + ll) // 2, ncols))
    _firdn_matrix(image, h, filtered)
    return filtered

cpdef upfir(double[:, :] image, double[:] h):
    """
    Upsamples the columns of the input image by 2, then applies the
    convolution kernel 'h' (again, columnwise). This is a special case of the
    matlab's 'upfirdn' function, ported to python. Returns the filtered image.

    Parameters
    ----------
    image: 2D array of doubles
        the input image to be filtered
    h: double array
        the convolution kernel
    """
    nrows = image.shape[0]
    ncols = image.shape[1]
    ll = h.shape[0]
    cdef double[:, :] filtered = np.zeros(shape=(2 * nrows + ll - 2, ncols))
    _upfir_matrix(image, h, filtered)
    return filtered


cdef void _average_block(double[:, :, :] image, int center_y, int center_x, int center_z,
                         double[:, :, :] weighted_average, double weight) noexcept nogil:
    """
    Compute the weighted average of the patches in a blockwise manner.

    This function accumulates weighted squared intensities from a block centered
    at (center_y, center_x, center_z) into the weighted_average array.

    Parameters
    ----------
    image : 3D array of doubles, shape (height, width, depth)
        Input image with standard array indexing [y, x, z]
    center_y : int
        Y coordinate (row index) of the center voxel
    center_x : int
        X coordinate (column index) of the center voxel
    center_z : int
        Z coordinate (depth index) of the center voxel
    weighted_average : 3D array of doubles
        Accumulator for weighted averages
    weight : double
        Weight for the weighted averaging
    """
    cdef int offset_y, offset_x, offset_z
    cdef int voxel_y, voxel_x, voxel_z
    cdef int is_outside_bounds
    cdef int block_radius = weighted_average.shape[0] // 2
    cdef int img_height = image.shape[0]
    cdef int img_width = image.shape[1]
    cdef int img_depth = image.shape[2]

    for offset_y in range(weighted_average.shape[0]):
        for offset_x in range(weighted_average.shape[1]):
            for offset_z in range(weighted_average.shape[2]):
                voxel_y = center_y + offset_y - block_radius
                voxel_x = center_x + offset_x - block_radius
                voxel_z = center_z + offset_z - block_radius

                is_outside_bounds = 0
                if voxel_y < 0 or voxel_y >= img_height:
                    is_outside_bounds = 1
                if voxel_x < 0 or voxel_x >= img_width:
                    is_outside_bounds = 1
                if voxel_z < 0 or voxel_z >= img_depth:
                    is_outside_bounds = 1

                if is_outside_bounds == 1:
                    # Use center voxel value for out-of-bounds locations
                    weighted_average[offset_y, offset_x, offset_z] += weight * (image[center_y, center_x, center_z]**2)
                else:
                    weighted_average[offset_y, offset_x, offset_z] += weight * (image[voxel_y, voxel_x, voxel_z]**2)


cdef void _value_block(double[:, :, :] denoised_estimate, double[:, :, :] weight_count,
                       int center_y, int center_x, int center_z,
                       double[:, :, :] weighted_average, double total_weight,
                       double noise_variance_doubled, int is_rician) noexcept nogil:
    """
    Computes the final denoised estimate for a block and accumulates it.

    This function processes each voxel in a block centered at (center_y, center_x, center_z),
    computes the denoised value from weighted averages, and accumulates the results.

    Parameters
    ----------
    denoised_estimate : 3D array of doubles, shape (height, width, depth)
        Accumulator for denoised intensity estimates
    weight_count : 3D array of doubles, shape (height, width, depth)
        Count of how many times each voxel has been processed
    center_y : int
        Y coordinate (row index) of the block center
    center_x : int
        X coordinate (column index) of the block center
    center_z : int
        Z coordinate (depth index) of the block center
    weighted_average : 3D array of doubles
        Weighted average intensities for the block
    total_weight : double
        Sum of all weights used in the averaging
    noise_variance_doubled : double
        2 * sigma^2 for Rician noise correction
    is_rician : int
        1 if Rician noise model, 0 for Gaussian
    """
    cdef int offset_y, offset_x, offset_z
    cdef int voxel_y, voxel_x, voxel_z
    cdef int is_outside_bounds
    cdef int block_radius = weighted_average.shape[0] // 2
    cdef int img_height = denoised_estimate.shape[0]
    cdef int img_width = denoised_estimate.shape[1]
    cdef int img_depth = denoised_estimate.shape[2]
    cdef double current_estimate, denoised_intensity, current_count

    for offset_y in range(weighted_average.shape[0]):
        for offset_x in range(weighted_average.shape[1]):
            for offset_z in range(weighted_average.shape[2]):
                voxel_y = center_y + offset_y - block_radius
                voxel_x = center_x + offset_x - block_radius
                voxel_z = center_z + offset_z - block_radius

                is_outside_bounds = 0
                if voxel_y < 0 or voxel_y >= img_height:
                    is_outside_bounds = 1
                if voxel_x < 0 or voxel_x >= img_width:
                    is_outside_bounds = 1
                if voxel_z < 0 or voxel_z >= img_depth:
                    is_outside_bounds = 1

                if is_outside_bounds == 0:
                    current_estimate = denoised_estimate[voxel_y, voxel_x, voxel_z]

                    # Compute denoised intensity from weighted average
                    if total_weight > 0:
                        if is_rician:
                            denoised_intensity = (weighted_average[offset_y, offset_x, offset_z] / total_weight) - noise_variance_doubled
                        else:
                            denoised_intensity = weighted_average[offset_y, offset_x, offset_z] / total_weight

                        if denoised_intensity > 0:
                            denoised_intensity = sqrt(denoised_intensity)
                        else:
                            denoised_intensity = 0.0
                    else:
                        denoised_intensity = 0.0

                    # Accumulate results
                    current_count = weight_count[voxel_y, voxel_x, voxel_z]
                    denoised_estimate[voxel_y, voxel_x, voxel_z] = current_estimate + denoised_intensity
                    weight_count[voxel_y, voxel_x, voxel_z] = current_count + 1.0


cdef double _patch_distance(double[:, :, :] image,
                           int patch1_center_y, int patch1_center_x, int patch1_center_z,
                           int patch2_center_y, int patch2_center_x, int patch2_center_z,
                           int patch_radius) nogil:
    """
    Computes the squared distance between two cubic patches in the image.

    Uses reflection padding at boundaries to handle edge cases properly.
    The distance is normalized by the number of voxels compared.

    Parameters
    ----------
    image : 3D array of doubles, shape (height, width, depth)
        Input image with standard array indexing [y, x, z]
    patch1_center_y, patch1_center_x, patch1_center_z : int
        Y, X, Z coordinates of the first patch center
    patch2_center_y, patch2_center_x, patch2_center_z : int
        Y, X, Z coordinates of the second patch center
    patch_radius : int
        Radius of the cubic patches (patch size = 2*radius + 1)

    Returns
    -------
    double
        Mean squared distance between the two patches
    """
    cdef double squared_distance_sum = 0.0
    cdef int voxel_count = 0
    cdef int offset_y, offset_x, offset_z
    cdef int voxel1_y, voxel1_x, voxel1_z
    cdef int voxel2_y, voxel2_x, voxel2_z
    cdef int img_height = image.shape[0]
    cdef int img_width = image.shape[1]
    cdef int img_depth = image.shape[2]
    cdef double intensity_diff

    for offset_y in range(-patch_radius, patch_radius + 1):
        for offset_x in range(-patch_radius, patch_radius + 1):
            for offset_z in range(-patch_radius, patch_radius + 1):
                # Calculate voxel coordinates for both patches
                voxel1_y = patch1_center_y + offset_y
                voxel1_x = patch1_center_x + offset_x
                voxel1_z = patch1_center_z + offset_z

                voxel2_y = patch2_center_y + offset_y
                voxel2_x = patch2_center_x + offset_x
                voxel2_z = patch2_center_z + offset_z

                # Apply reflection padding for out-of-bounds coordinates
                # Y dimension (height)
                if voxel1_y < 0:
                    voxel1_y = -voxel1_y
                elif voxel1_y >= img_height:
                    voxel1_y = 2 * img_height - voxel1_y - 1

                if voxel2_y < 0:
                    voxel2_y = -voxel2_y
                elif voxel2_y >= img_height:
                    voxel2_y = 2 * img_height - voxel2_y - 1

                # X dimension (width)
                if voxel1_x < 0:
                    voxel1_x = -voxel1_x
                elif voxel1_x >= img_width:
                    voxel1_x = 2 * img_width - voxel1_x - 1

                if voxel2_x < 0:
                    voxel2_x = -voxel2_x
                elif voxel2_x >= img_width:
                    voxel2_x = 2 * img_width - voxel2_x - 1

                # Z dimension (depth)
                if voxel1_z < 0:
                    voxel1_z = -voxel1_z
                elif voxel1_z >= img_depth:
                    voxel1_z = 2 * img_depth - voxel1_z - 1

                if voxel2_z < 0:
                    voxel2_z = -voxel2_z
                elif voxel2_z >= img_depth:
                    voxel2_z = 2 * img_depth - voxel2_z - 1

                # Compute squared difference
                intensity_diff = image[voxel1_y, voxel1_x, voxel1_z] - image[voxel2_y, voxel2_x, voxel2_z]
                squared_distance_sum += intensity_diff * intensity_diff
                voxel_count += 1

    return squared_distance_sum / voxel_count


cdef void _process_block_complete(double[:, :, :] image,
                                double[:, :, :] mask,
                                double[:, :, :] local_means,
                                double[:, :, :] local_variances,
                                double[:, :, :] accumulated_estimates,
                                double[:, :, :] weight_counts,
                                double[:, :, :] workspace,
                                int center_y, int center_x, int center_z,
                                int patch_radius, int block_radius,
                                double filtering_strength,
                                double noise_variance_doubled, int is_rician,
                                double epsilon, double mean_ratio_threshold,
                                double variance_ratio_min,
                                int img_height, int img_width, int img_depth) noexcept nogil:
    """
    Process a complete block including weight computation and final application.

    This function avoids reduction variables by computing and applying weights
    in a single pass without intermediate storage.
    """
    cdef double current_mean = local_means[center_y, center_x, center_z]
    cdef double current_variance = local_variances[center_y, center_x, center_z]
    cdef double accumulator_weight = 0.0
    cdef double max_weight = 0.0
    cdef double similarity_weight, patch_distance, mean_ratio, variance_ratio
    cdef double neighbor_mean, neighbor_variance
    cdef int neighbor_y, neighbor_x, neighbor_z
    cdef int offset_y, offset_x, offset_z
    cdef double variance_ratio_max = 1.0 / variance_ratio_min

    # Handle low signal/noise regions with simple averaging
    if current_mean <= epsilon or current_variance <= epsilon:
        max_weight = 1.0
        _average_block(image, center_y, center_x, center_z, workspace, max_weight)
        accumulator_weight = max_weight
    else:
        # Search for similar patches in neighborhood
        for offset_z in range(-patch_radius, patch_radius + 1):
            neighbor_z = center_z + offset_z
            if neighbor_z < 0 or neighbor_z >= img_depth:
                continue

            for offset_x in range(-patch_radius, patch_radius + 1):
                neighbor_x = center_x + offset_x
                if neighbor_x < 0 or neighbor_x >= img_width:
                    continue

                for offset_y in range(-patch_radius, patch_radius + 1):
                    neighbor_y = center_y + offset_y
                    if neighbor_y < 0 or neighbor_y >= img_height:
                        continue
                    if neighbor_y == center_y and neighbor_x == center_x and neighbor_z == center_z:
                        continue
                    if mask[neighbor_y, neighbor_x, neighbor_z] == 0:
                        continue

                    neighbor_mean = local_means[neighbor_y, neighbor_x, neighbor_z]
                    neighbor_variance = local_variances[neighbor_y, neighbor_x, neighbor_z]

                    # Skip patches with insufficient signal
                    if neighbor_mean <= epsilon or neighbor_variance <= epsilon:
                        continue

                    # Pre-filtering based on statistical similarity
                    mean_ratio = current_mean / neighbor_mean
                    variance_ratio = current_variance / neighbor_variance

                    if (mean_ratio < mean_ratio_threshold or mean_ratio > (1.0 / mean_ratio_threshold) or
                        variance_ratio < variance_ratio_min or variance_ratio > variance_ratio_max):
                        continue

                    # Compute patch distance
                    patch_distance = _patch_distance(image,
                                                    center_y, center_x, center_z,
                                                    neighbor_y, neighbor_x, neighbor_z,
                                                    block_radius)

                    # Compute similarity weight using original nlmeans formula
                    # Original: w = exp(-d / (h * h)) where h is the noise sigma
                    similarity_weight = exp(-patch_distance / filtering_strength)

                    if similarity_weight > max_weight:
                        max_weight = similarity_weight

                    # Accumulate weighted averages directly
                    _average_block(image, neighbor_y, neighbor_x, neighbor_z, workspace, similarity_weight)
                    accumulator_weight += similarity_weight

    # Apply accumulated weighted averages to final estimate
    if accumulator_weight > 0.0:
        _value_block(accumulated_estimates, weight_counts,
                   center_y, center_x, center_z, workspace,
                   accumulator_weight, noise_variance_doubled, is_rician)


cdef inline void _clear_workspace(double[:, :, :] workspace) noexcept nogil:
    """
    Efficiently clears a 3D workspace array by setting all values to zero.

    Parameters
    ----------
    workspace : 3D array of doubles
        The workspace array to clear
    """
    cdef int i, j, k
    cdef int size_i = workspace.shape[0]
    cdef int size_j = workspace.shape[1]
    cdef int size_k = workspace.shape[2]

    for i in range(size_i):
        for j in range(size_j):
            for k in range(size_k):
                workspace[i, j, k] = 0.0


cdef double _local_mean(double[:, :, :] image, int center_y, int center_x, int center_z, int patch_radius) nogil:
    """
    Computes the local mean of a cubic patch centered at (center_y, center_x, center_z).

    Uses reflection padding for boundary handling.

    Parameters
    ----------
    image : 3D array of doubles, shape (height, width, depth)
        Input image with standard array indexing [y, x, z]
    center_y, center_x, center_z : int
        Y, X, Z coordinates of the patch center
    patch_radius : int
        Radius of the patch (patch size = 2*radius + 1 in each dimension)

    Returns
    -------
    double
        Mean intensity value of the patch
    """
    cdef int img_height = image.shape[0]
    cdef int img_width = image.shape[1]
    cdef int img_depth = image.shape[2]
    cdef double intensity_sum = 0.0
    cdef int voxel_count = 0
    cdef int voxel_y, voxel_x, voxel_z
    cdef int offset_y, offset_x, offset_z

    for offset_y in range(-patch_radius, patch_radius + 1):
        for offset_x in range(-patch_radius, patch_radius + 1):
            for offset_z in range(-patch_radius, patch_radius + 1):
                voxel_y = center_y + offset_y
                voxel_x = center_x + offset_x
                voxel_z = center_z + offset_z

                # Apply reflection padding
                if voxel_y < 0:
                    voxel_y = -voxel_y
                elif voxel_y >= img_height:
                    voxel_y = 2 * img_height - voxel_y - 1

                if voxel_x < 0:
                    voxel_x = -voxel_x
                elif voxel_x >= img_width:
                    voxel_x = 2 * img_width - voxel_x - 1

                if voxel_z < 0:
                    voxel_z = -voxel_z
                elif voxel_z >= img_depth:
                    voxel_z = 2 * img_depth - voxel_z - 1

                intensity_sum += image[voxel_y, voxel_x, voxel_z]
                voxel_count += 1

    return intensity_sum / voxel_count


cdef double _local_variance(double[:, :, :] image, double mean_intensity,
                           int center_y, int center_x, int center_z, int patch_radius) nogil:
    """
    Computes the local variance of a cubic patch centered at (center_y, center_x, center_z).

    Uses reflection padding for boundary handling.

    Parameters
    ----------
    image : 3D array of doubles, shape (height, width, depth)
        Input image with standard array indexing [y, x, z]
    mean_intensity : double
        Pre-computed mean intensity of the patch
    center_y, center_x, center_z : int
        Y, X, Z coordinates of the patch center
    patch_radius : int
        Radius of the patch (patch size = 2*radius + 1 in each dimension)

    Returns
    -------
    double
        Variance of intensities in the patch
    """
    cdef int img_height = image.shape[0]
    cdef int img_width = image.shape[1]
    cdef int img_depth = image.shape[2]
    cdef double squared_diff_sum = 0.0
    cdef int voxel_count = 0
    cdef int voxel_y, voxel_x, voxel_z
    cdef int offset_y, offset_x, offset_z
    cdef double intensity_diff

    for offset_y in range(-patch_radius, patch_radius + 1):
        for offset_x in range(-patch_radius, patch_radius + 1):
            for offset_z in range(-patch_radius, patch_radius + 1):
                voxel_y = center_y + offset_y
                voxel_x = center_x + offset_x
                voxel_z = center_z + offset_z

                # Apply reflection padding
                if voxel_y < 0:
                    voxel_y = -voxel_y
                elif voxel_y >= img_height:
                    voxel_y = 2 * img_height - voxel_y - 1

                if voxel_x < 0:
                    voxel_x = -voxel_x
                elif voxel_x >= img_width:
                    voxel_x = 2 * img_width - voxel_x - 1

                if voxel_z < 0:
                    voxel_z = -voxel_z
                elif voxel_z >= img_depth:
                    voxel_z = 2 * img_depth - voxel_z - 1

                intensity_diff = image[voxel_y, voxel_x, voxel_z] - mean_intensity
                squared_diff_sum += intensity_diff * intensity_diff
                voxel_count += 1

    # Use N-1 for sample variance (Bessel's correction)
    if voxel_count > 1:
        return squared_diff_sum / (voxel_count - 1)
    else:
        return 0.0


def nlmeans_3d_blockwise(double[:, :, :] image, double[:, :, :] mask, int patch_radius, int block_radius, noise_sigma, int is_rician, num_threads=None):
    """
    Non-Local Means Denoising Using Blockwise Averaging.

    This implementation uses efficient memory management and parallel processing
    for high-performance denoising of 3D images.

    See :footcite:p:`Coupe2008` and :footcite:p:`Coupe2012` for further details
    about the method.

    Parameters
    ----------
    image : 3D array of doubles, shape (height, width, depth)
        Input image corrupted with noise. Uses standard array indexing [y, x, z].
    mask : 3D array of doubles, shape (height, width, depth)
        Binary mask indicating which voxels to process (1) or skip (0).
    patch_radius : int
        Radius for local patch comparisons. Search is performed within a
        (2*patch_radius + 1)^3 neighborhood around each voxel.
    block_radius : int
        Radius for the blocks used in weighted averaging. Each block has
        size (2*block_radius + 1)^3.
    noise_sigma : double or 3D ndarray
        Estimated noise standard deviation in the input image. Can be a
        scalar (uniform noise level) or a 3D sigma map with shape matching
        ``image.shape``.
    is_rician : int
        1 if Rician noise model should be used, 0 for Gaussian noise model.
    num_threads : int, optional
        Number of OpenMP threads to use. If None, uses all available threads.

    Returns
    -------
    denoised_image : 3D array of doubles
        Denoised output image with the same shape as input.

    References
    ----------
    .. footbibliography::
    """
    # Image dimensions
    cdef int img_height = image.shape[0]
    cdef int img_width = image.shape[1]
    cdef int img_depth = image.shape[2]

    # Handle num_threads parameter
    cdef int threads_to_use
    if num_threads is None:
        threads_to_use = determine_num_threads(None)
    else:
        threads_to_use = determine_num_threads(num_threads)

    # Algorithm parameters
    cdef bint use_sigma_map = False
    cdef double sigma_scalar = 0.0
    cdef double[:, :, :] sigma_map = None
    cdef double current_sigma
    cdef double current_sigma_sq
    cdef double local_noise_variance_doubled
    cdef double local_filtering_strength

    if hasattr(noise_sigma, 'ndim'):
        if noise_sigma.ndim != 3:
            raise ValueError(
                f'noise_sigma should be scalar or a 3D ndarray, '
                f'got {getattr(noise_sigma, "shape", type(noise_sigma))}'
            )
        if noise_sigma.shape != (img_height, img_width, img_depth):
            raise ValueError(
                f'3D noise_sigma shape {noise_sigma.shape} does not match '
                f'image shape ({img_height}, {img_width}, {img_depth})'
            )
        sigma_map = np.ascontiguousarray(noise_sigma, dtype='f8')
        use_sigma_map = True
    else:
        sigma_scalar = float(noise_sigma)

    cdef int block_size = 2 * block_radius + 1

    # Statistical filtering thresholds
    cdef double epsilon = 1e-5
    cdef double mean_ratio_threshold = 0.95
    cdef double variance_ratio_min = 0.5 + 1e-7
    cdef double variance_ratio_max = 1.0 / variance_ratio_min

    # Output arrays
    cdef double[:, :, :] denoised_image = np.zeros_like(image)
    cdef double[:, :, :] local_means = np.zeros_like(image)
    cdef double[:, :, :] local_variances = np.zeros_like(image)
    cdef double[:, :, :] accumulated_estimates = np.zeros_like(image)
    cdef double[:, :, :] weight_counts = np.zeros_like(image)

    # Thread-local working memory (allocated per thread to avoid race conditions)
    cdef double *thread_weighted_averages = NULL
    cdef int block_volume = block_size * block_size * block_size

    # Loop variables
    cdef int center_y, center_x, center_z
    cdef int neighbor_y, neighbor_x, neighbor_z
    cdef int offset_y, offset_x, offset_z
    cdef double mean_ratio, variance_ratio
    cdef double patch_distance, similarity_weight, max_weight, total_weight
    cdef double current_mean, current_variance
    cdef double neighbor_mean, neighbor_variance
    cdef int thread_id

    # Set number of threads
    set_num_threads(threads_to_use)

    # Pre-allocate weighted average workspace (one per potential thread)
    cdef int max_threads = threads_to_use
    cdef double[:, :, :, :] thread_workspaces = np.zeros((max_threads, block_size, block_size, block_size), dtype=np.float64)

    # Phase 1: Compute local statistics (means and variances) in parallel
    with nogil, parallel():
        for center_z in prange(img_depth, schedule="dynamic"):
            for center_x in range(img_width):
                for center_y in range(img_height):
                    if mask[center_y, center_x, center_z] > 0:
                        current_mean = _local_mean(image, center_y, center_x, center_z, 1)  # 3x3x3 patch
                        local_means[center_y, center_x, center_z] = current_mean
                        local_variances[center_y, center_x, center_z] = _local_variance(
                            image, current_mean, center_y, center_x, center_z, 1)

    # Phase 2: Process blocks with stride 2 for efficiency
    # Avoid reduction variables entirely by processing each block completely.
    # Two separate code paths avoid a per-voxel branch on use_sigma_map.
    cdef double fixed_sigma_sq = sigma_scalar * sigma_scalar
    if fixed_sigma_sq <= 0:
        fixed_sigma_sq = epsilon
    cdef double fixed_noise_variance_doubled = 2.0 * fixed_sigma_sq
    cdef double fixed_filtering_strength = fixed_sigma_sq

    if use_sigma_map:
        with nogil, parallel():
            for center_z in prange(0, img_depth, 2, schedule="dynamic"):
                thread_id = threadid()
                for center_x in range(0, img_width, 2):
                    for center_y in range(0, img_height, 2):
                        if mask[center_y, center_x, center_z] == 0:
                            continue

                        current_mean = local_means[center_y, center_x, center_z]
                        current_variance = local_variances[center_y, center_x, center_z]

                        _clear_workspace(thread_workspaces[thread_id])

                        current_sigma = sigma_map[center_y, center_x, center_z]
                        current_sigma_sq = current_sigma * current_sigma
                        if current_sigma_sq <= 0:
                            current_sigma_sq = epsilon
                        local_noise_variance_doubled = 2.0 * current_sigma_sq
                        local_filtering_strength = current_sigma_sq

                        _process_block_complete(image, mask, local_means, local_variances,
                                              accumulated_estimates, weight_counts,
                                              thread_workspaces[thread_id],
                                              center_y, center_x, center_z,
                                              patch_radius, block_radius, local_filtering_strength,
                                              local_noise_variance_doubled, is_rician,
                                              epsilon, mean_ratio_threshold, variance_ratio_min,
                                              img_height, img_width, img_depth)
    else:
        with nogil, parallel():
            for center_z in prange(0, img_depth, 2, schedule="dynamic"):
                thread_id = threadid()
                for center_x in range(0, img_width, 2):
                    for center_y in range(0, img_height, 2):
                        if mask[center_y, center_x, center_z] == 0:
                            continue

                        current_mean = local_means[center_y, center_x, center_z]
                        current_variance = local_variances[center_y, center_x, center_z]

                        _clear_workspace(thread_workspaces[thread_id])

                        _process_block_complete(image, mask, local_means, local_variances,
                                              accumulated_estimates, weight_counts,
                                              thread_workspaces[thread_id],
                                              center_y, center_x, center_z,
                                              patch_radius, block_radius, fixed_filtering_strength,
                                              fixed_noise_variance_doubled, is_rician,
                                              epsilon, mean_ratio_threshold, variance_ratio_min,
                                              img_height, img_width, img_depth)

    # Phase 3: Finalize denoised estimates
    with nogil, parallel():
        for center_z in prange(img_depth, schedule="static"):
            for center_x in range(img_width):
                for center_y in range(img_height):
                    if mask[center_y, center_x, center_z] == 0:
                        denoised_image[center_y, center_x, center_z] = 0.0
                    elif weight_counts[center_y, center_x, center_z] > 0:
                        denoised_image[center_y, center_x, center_z] = (
                            accumulated_estimates[center_y, center_x, center_z] /
                            weight_counts[center_y, center_x, center_z])
                    else:
                        # Fallback to original value if no similar patches found
                        denoised_image[center_y, center_x, center_z] = image[center_y, center_x, center_z]

    # Restore default thread count
    restore_default_num_threads()

    return np.asarray(denoised_image)


def nlmeans_3d(arr, mask=None, sigma=None, patch_radius=1,
               block_radius=5, rician=True, num_threads=None, method='classic'):
    """ Non-local means for denoising 3D images with selectable algorithm

    Parameters
    ----------
    arr : 3D ndarray
        The array to be denoised
    mask : 3D ndarray
    sigma : float or 3D array
        standard deviation of the noise estimated from the data
    patch_radius : int
        patch size is ``2 x patch_radius + 1``. Default is 1.
    block_radius : int
        block size is ``2 x block_radius + 1``. Default is 5.
    rician : boolean
        If True the noise is estimated as Rician, otherwise Gaussian noise
        is assumed.
    num_threads : int, optional
        Number of threads to be used for OpenMP parallelization. If None
        (default) the value of OMP_NUM_THREADS environment variable is used
        if it is set, otherwise all available threads are used. If < 0 the
        maximal number of threads minus $|num_threads + 1|$ is used (enter -1 to
        use as many threads as possible). 0 raises an error.
    method : str, optional
        Algorithm method to use. 'classic' for original algorithm,
        'blockwise' for improved blockwise algorithm. Default is 'classic'.

    Returns
    -------
    denoised_arr : ndarray
        the denoised ``arr`` which has the same shape as ``arr``.
    """

    if arr.ndim != 3:
        raise ValueError(f'data needs to be a 3D ndarray, got shape {arr.shape}')

    if mask is None:
        mask = np.ones(arr.shape, dtype='f8')
    else:
        mask = np.ascontiguousarray(mask, dtype='f8')

    if mask.ndim != 3:
        raise ValueError(f'mask needs to be a 3D ndarray, got shape {mask.shape}')

    # Handle sigma validation based on method
    if method == 'classic':
        # Classic method requires 3D sigma array
        if not hasattr(sigma, 'ndim') or sigma.ndim != 3:
            raise ValueError(
                f'sigma needs to be a 3D ndarray for classic method, '
                f'got {getattr(sigma, "shape", type(sigma))}'
            )
    elif method == 'blockwise':
        # Blockwise method can accept scalar or array sigma
        if hasattr(sigma, 'ndim'):
            if sigma.ndim > 3:
                raise ValueError(
                    f'sigma should be at most 3D for blockwise method, '
                    f'got shape {sigma.shape}'
                )
        # Scalar sigma is fine for blockwise method
    else:
        raise ValueError(f"Unknown method '{method}'. Use 'classic' or 'blockwise'.")

    arr = np.ascontiguousarray(arr, dtype='f8')

    if method == 'classic':
        # Use original classic algorithm with padding
        arr = add_padding_reflection(arr, block_radius)
        mask = add_padding_reflection(mask.astype('f8'), block_radius)
        sigma = np.ascontiguousarray(sigma, dtype='f8')
        sigma = add_padding_reflection(sigma.astype('f8'), block_radius)
        arrnlm = _nlmeans_3d(arr, mask, sigma, patch_radius, block_radius,
                             rician, num_threads)
        return remove_padding(arrnlm, block_radius)
    elif method == 'blockwise':
        # Use new blockwise algorithm without padding
        # For blockwise, scalar or 3D sigma map are supported
        if hasattr(sigma, 'shape'):
            if sigma.shape == arr.shape:
                sigma_input = np.ascontiguousarray(sigma, dtype='f8')
            elif sigma.ndim == 1:
                # 1D sigma for 3D input: take the mean as a fallback.
                # For 4D inputs, nlmeans.py extracts per-volume scalars before reaching here.
                sigma_input = float(np.mean(sigma))
            elif sigma.shape == ():
                # 0-D array (scalar in array form)
                sigma_input = float(sigma)
            else:
                raise ValueError(f'Invalid sigma shape {sigma.shape} for blockwise method')
        else:
            # Scalar sigma
            sigma_input = float(sigma)

        return nlmeans_3d_blockwise(arr, mask, patch_radius, block_radius,
                                    sigma_input, int(rician), num_threads)
    else:
        raise ValueError(f"Unknown method '{method}'. Use 'classic' or 'blockwise'.")



"""
=====================================================================
Robust fitting with iteratively reweighted least squares.
=====================================================================


Both :footcite:t:`Coveney2025` and :footcite:t:`Coveney2025b` show
robust fitting using iteretively reweighted least squares (IRLS).
Methods for fitting DTI and DKI data are shown in the relevant examples.

It is possible to construct novel weighting functions and novel outlier
functions in DIPY_ in order to do robust fitting with IRLS.

Here we show an example of multi-voxel outlier detection (MVOD)
:footcite:t:`Coveney2025`.
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti
import dipy.reconst.dti as dti
from dipy.reconst.weights_method import (
    simple_cutoff,
    weights_method_wls_m_est,
)
from dipy.segment.mask import median_otsu
from dipy.utils.volume import adjacency_calc
from dipy.viz.plotting import compare_maps

###############################################################################
# We will use the same data from the
# :ref:`DTI example <sphx_glr_examples_built_reconstruction_reconst_dti.py>`.

hardi_fname, hardi_bval_fname, hardi_bvec_fname = get_fnames(name="stanford_hardi")

data, affine = load_nifti(hardi_fname)

bvals, bvecs = read_bvals_bvecs(hardi_bval_fname, hardi_bvec_fname)
gtab = gradient_table(bvals, bvecs=bvecs)


maskdata, mask = median_otsu(
    data, vol_idx=range(10, 50), median_radius=3, numpass=1, autocrop=True, dilate=2
)

###############################################################################
# Firstly let's fit with WLS.

tenmodel = dti.TensorModel(gtab, fit_method="WLS")
tenfit_orig = tenmodel.fit(maskdata[:, :, 38:39], mask=mask[:, :, 38:39])

###############################################################################
# Examples :ref:`DTI <sphx_glr_examples_built_reconstruction_reconst_dti.py>`
# and :ref:`DKI <sphx_glr_examples_built_reconstruction_reconst_dki.py>`
# both show how to use robust fitting.
# Let's corrupt the data by off-setting a bunch of images.
# Of course, in this case shot rejection could be used,
# but this is just an easy way to make an example.
# Let's see how WLS and Robust WLS (RWLS) perform.

maskdata[30:, 30:, 38:39, 10:20] = maskdata[0:-30, 0:-30, 38:39, 10:20]

tenmodel = dti.TensorModel(gtab, fit_method="WLS")
tenfit = tenmodel.fit(maskdata[:, :, 38:39], mask=mask[:, :, 38:39])

tenmodel_rwls = dti.TensorModel(gtab, fit_method="RWLS")
tenfit_rwls = tenmodel_rwls.fit(maskdata[:, :, 38:39], mask=mask[:, :, 38:39])

###############################################################################
# Here are MD and FA for the different fits:

compare_maps(
    [tenfit_orig, tenfit, tenfit_rwls],
    ["md", "fa"],
    fit_labels=["WLS orig", "WLS", "RWLS"],
    map_kwargs=[{"vmin": 0.000, "vmax": 0.002}, {"vmin": 0, "vmax": 1.0}],
    filename="Compare_WLS_and_RWLS.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# WLS and RWLS on the corrupted data, compared to the original fit.
#
#
#
# We might be able to do better by designing our own weight function to use on
# each iteration of IRLS, as well as by designing our own outlier rejection
# functions.
#
# Here, for simplicity, we will load the
# :func:`~dipy.reconst.weights_method.weights_method_wls_m_est`
# function (take a look at this function to understand how you might write your
# own), but write our own outlier rejection function that depends on
# neighbouring voxels.
#
# Firstly, let's calculate adjacency, respecting the mask. For each voxel, we
# obtain the index (into a masked array of neighbouring voxels), with
# voxel-distance less than ``cutoff``.

adjacency = adjacency_calc(mask.shape[0:2], mask=mask[:, :, 38:39], cutoff=3)

###############################################################################
# Now we will write a multi-voxel outlier detection (MVOD) function.


def MVOD_cutoff(
    residuals,
    log_residuals,
    pred_sig,
    design_matrix,
    leverages,
    C,
    cutoff,
    linear,
    adjacency,
):
    """MVOD outlier detection (includes "two-eyes" SVOD condition as well)"""

    # SVOD: single voxel outlier detection
    cond = simple_cutoff(
        residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff
    )
    # an alternative outlier definition for single voxels is:
    # cond = two_eyes_cutoff(residuals, log_residuals, pred_sig,
    #                        design_matrix, leverages, C, cutoff)

    # MVOD: multiple voxel outlier detection
    # ======================================
    MV_cond = np.zeros_like(cond)
    leverages[np.isclose(leverages, 1.0)] = 0.9999
    HAT_factor = np.sqrt(1 - leverages)
    _p = design_matrix.shape[1]
    factor = 1.4826

    # if adjacency is None, then weights are the same for all voxels
    adj_full = False
    if adjacency is None:
        adjacency = [[list(range(residuals.shape[0]))]]
        adj_full = True

    for vox in range(len(adjacency)):
        if linear:
            HAT, _HAT_vox = HAT_factor[adjacency[vox]], HAT_factor[vox]
        else:
            HAT, _HAT_vox = HAT_factor, HAT_factor

        # calculate C from all adjacent voxels
        residuals_vox = residuals[adjacency[vox]]
        _log_residuals_vox = log_residuals[adjacency[vox]]
        _pred_sig_vox = pred_sig[adjacency[vox]]

        yy = residuals_vox / HAT
        # C is a robust estimate of stdev of error
        # here, uses all images and all voxels in adjacency
        C = factor * np.median(np.abs(yy - np.median(yy)))

        # multivariate outlier condition: RMSE
        MV_cutoff = cutoff
        RMSE = np.sqrt((yy**2).mean(axis=0))
        MED = np.median(RMSE)
        MAD = 1.4826 * np.median(np.abs(RMSE - MED))
        cond_RMSE = RMSE > (MED + MV_cutoff * MAD)
        MV_cond[vox] = MV_cond[vox] | cond_RMSE

        # multivariate outlier condition: SE
        MV_cutoff = cutoff
        SE = np.sum(yy, axis=0)  # sum of residuals
        MED = np.median(SE)
        MAD = 1.4826 * np.median(np.abs(SE - MED))
        cond_SE = (SE > (MED + MV_cutoff * MAD)) | (SE < (MED - MV_cutoff * MAD))
        MV_cond[vox] = MV_cond[vox] | cond_SE

    if adj_full:
        MV_cond[1:] = MV_cond[0]

    # check condition number that will result
    for vox in range(cond.shape[0]):
        u, s, vh = np.linalg.svd(design_matrix[np.invert(MV_cond[vox] | cond[vox])])
        min_s = s.min()
        if min_s < 1e-2:
            pass
        else:
            # include MV_cond if it doesn't make the design_matrix singular
            cond[vox] = cond[vox] | MV_cond[vox]

    return cond


###############################################################################
# Outlier condition functions in DIPY_ must only take the following arguments:
# ```residuals, log_residuals, pred_sig, design_matrix, leverages, C, cutoff```
#
# This is so that weighting functions, which make calls to outlier condition
# functions, are always able to pass the required arguments to the outlier
# functions.
#
# To handle the fact that our outlier function depends on two more arguments,
# we create a function as follows (adjacency is in scope here):


def mvod(*args):
    return MVOD_cutoff(*args, linear=True, adjacency=adjacency)


###############################################################################
# Now we are free to define our weights function. Even though we will use
# ``weights_method_wls_m_est``, we will specify a new weights function that
# defines the outlier condition, and other additional arguments, as follows:


def wm(*args):
    return weights_method_wls_m_est(
        *args, m_est="gm", cutoff=3, outlier_condition_func=mvod
    )


###############################################################################
# This is because weights functions in DIPY_ must only take arguments
# ``data, pred_sig, design_matrix, leverages, rdx, TDX, robust``
# (and must return only ``w, robust``).

tenmodel_rwls_mvod = dti.TensorModel(
    gtab,
    fit_method="IRLS",
    return_S0_hat=True,
    weights_method=wm,
    fit_type="WLS",
    num_iter=10,
)
tenfit_rwls_mvod = tenmodel_rwls_mvod.fit(maskdata[:, :, 38:39], mask=mask[:, :, 38:39])

###############################################################################
# Robustly fitting (via IRLS) with the new weights function gives much better
# results.

compare_maps(
    [tenfit_orig, tenfit, tenfit_rwls_mvod],
    ["md", "fa"],
    fit_labels=["WLS orig", "WLS", "RWLS + MVOD"],
    map_kwargs=[{"vmin": 0.000, "vmax": 0.002}, {"vmin": 0, "vmax": 1.0}],
    filename="Compare_WLS_and_RWLS_MVOD.png",
)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# WLS and RWLS with MVOD on the corrupted data, compared to the original fit.
#
#
# Of course you are free to write your own weights functions from scratch.
# Since weights are calculated after each iteration of fitting over all masked
# voxels, both the weights function and the outlier condition function can
# calculate weights and define outliers for the current voxels using
# information from other voxels.
#
#
#
#
# References
# ----------
#
# .. footbibliography::
#



import logging
import os
import sys


class CustomHandler(logging.Handler):
    """Custom logging handler that writes an empty line for empty log messages,
    otherwise formats the message as usual.
    """

    def __init__(self, stream=None, filename=None):
        super().__init__()  # call with no arguments
        if filename is not None:
            self._should_close = True
            self.stream = open(filename, "a", encoding="utf-8")
        else:
            self._should_close = False
            self.stream = stream if stream is not None else sys.stdout

    def emit(self, record):
        try:
            msg = record.getMessage()
            if msg == "":
                self.stream.write("\n")
            else:
                formatted = self.format(record)
                self.stream.write(formatted + "\n")
            self.flush()
        except Exception:
            self.handleError(record)

    def flush(self):
        if self.stream and hasattr(self.stream, "flush"):
            self.stream.flush()

    def close(self):
        if self._should_close:
            try:
                self.stream.close()
            except Exception:
                pass
        super().close()


def get_logger(name="dipy", filename=None, force=False):
    """Return a logger instance configured for DIPY.

    Parameters
    ----------
    name : str
        The logger name.
    filename : str, Path or None, optional
        If provided, log messages will also be saved to this file. If ``None``,
        logs are sent to stdout.
    force : bool, optional
        If True, existing handlers attached to the logger will be removed
        and replaced with a new handler. This allows reconfiguration of the
        logger even if it was already set up.

    Returns
    -------
    _logger : logging.Logger
        Configured logger.
    """

    _logger = logging.getLogger(name)
    if force or not _logger.hasHandlers():
        _logger.handlers.clear()
        if filename:
            handler = CustomHandler(filename=filename)
        else:
            handler = CustomHandler(stream=sys.stdout)
        handler.setLevel(logging.INFO)
        formatter = logging.Formatter(
            "[%(asctime)s][%(name)s] %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        handler.setFormatter(formatter)
        _logger.addHandler(handler)
        _logger.setLevel(logging.INFO)
        # Enable propagation when running tests so pytest caplog can capture messages
        # before we were using False to prevent propagation to root logger
        _logger.propagate = "PYTEST_CURRENT_TEST" in os.environ
    return _logger


def configure_logger(
    level=logging.INFO,
    fmt="[%(asctime)s][%(name)s] %(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    filename=None,
):
    """Reconfigure DIPY logger.

    Parameters
    ----------
    level : int
        Logging level (e.g., logging.INFO).
    fmt : str, optional
        Log message format.
    datefmt : str, optional
        Date format for log messages.
    filename : str, Path or None, optional
        If provided, log messages will also be saved to this file. If ``None``,
        logs are sent to stdout.
    """

    # Remove all handlers associated with the root logger object.
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)

    if filename:
        handler = CustomHandler(filename=filename)
    else:
        handler = CustomHandler(stream=sys.stdout)

    handler.setLevel(level)
    handler.setFormatter(logging.Formatter(fmt, datefmt=datefmt))
    logging.root.addHandler(handler)
    logging.root.setLevel(level)


# Provide a default logger for convenience
logger = get_logger()



import logging
from pathlib import Path
from tempfile import TemporaryDirectory
import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.core.gradients import generate_bvecs
from dipy.data import get_fnames
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.io.peaks import load_pam
from dipy.reconst.mcsd import have_cvxpy
from dipy.reconst.shm import descoteaux07_legacy_msg, sph_harm_ind_list
from dipy.testing import assert_warns
from dipy.workflows.reconst import ReconstCSDFlow, ReconstQBallBaseFlow, ReconstSDTFlow

needs_cvxpy = pytest.mark.skipif(not have_cvxpy, reason="Requires CVXPY")

logging.getLogger().setLevel(logging.INFO)


def test_reconst_csa():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstQBallBaseFlow)


def test_reconst_opdt():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstQBallBaseFlow, method="opdt")


def test_reconst_qball():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstQBallBaseFlow, method="qball")


def test_reconst_csd():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstCSDFlow)


def test_reconst_sdt():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstSDTFlow)


@needs_cvxpy
def test_reconst_csd_msmt():
    """Test MSMT-CSD functionality."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        warnings.filterwarnings(
            "ignore",
            message="Solution may be inaccurate.*",
            category=UserWarning,
        )
        with TemporaryDirectory() as out_dir:
            data_path, bval_path, bvec_path = get_fnames(name="small_64D")
            volume, affine = load_nifti(data_path)
            mask = np.ones_like(volume[:, :, :, 0])
            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.uint8), affine)

            reconst_flow = ReconstCSDFlow()
            reconst_flow._force_overwrite = True

            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                use_msmt=True,
                iso=3,
                sh_order_max=4,
                out_dir=out_dir,
                extract_pam_values=True,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            npt.assert_equal(load_nifti_data(gfa_path).shape, volume.shape[:-1])

            for fname in ["wm_mask.nii.gz", "gm_mask.nii.gz", "csf_mask.nii.gz"]:
                mask_path_out = Path(out_dir) / fname
                assert mask_path_out.exists()
                npt.assert_equal(
                    load_nifti_data(str(mask_path_out)).shape, volume.shape[:-1]
                )


@needs_cvxpy
def test_reconst_csd_msmt_with_t1():
    """Test MSMT-CSD with provided T1 image (T1-based HMRF tissue segmentation).

    The anisotropic power map is used as the T1 substitute so that the HMRF
    tissue labels are consistent with the actual DWI signal properties,
    avoiding degenerate response functions.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        warnings.filterwarnings(
            "ignore",
            message="Solution may be inaccurate.*",
            category=UserWarning,
        )
        with TemporaryDirectory() as out_dir:
            data_path, bval_path, bvec_path = get_fnames(name="small_64D")
            volume, affine = load_nifti(data_path)
            bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)
            mask = np.ones_like(volume[:, :, :, 0])
            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.uint8), affine)

            # Compute the anisotropic power map from DWI data and save it as a
            # "T1" so HMRF tissue labels are consistent with DWI signal.
            from dipy.core.gradients import gradient_table
            from dipy.core.sphere import HemiSphere
            from dipy.reconst.shm import (
                anisotropic_power,
                normalize_data,
                smooth_pinv,
                sph_harm_lookup,
            )

            gtab = gradient_table(bvals, bvecs=bvecs)
            dwi_mask = ~gtab.b0s_mask
            normed = normalize_data(volume, gtab.b0s_mask)[..., dwi_mask]
            normed = normed * mask[..., None]
            signal_pts = HemiSphere(xyz=gtab.bvecs[dwi_mask])
            Ba, m, n = sph_harm_lookup.get("descoteaux07")(
                4, signal_pts.theta, signal_pts.phi
            )
            invB = smooth_pinv(Ba, np.sqrt(0.0) * (-n * (n + 1)))
            shm = np.dot(normed, invB.T)
            amap = anisotropic_power(
                shm, norm_factor=0.00001, power=2, non_negative=True
            )
            t1_path = Path(out_dir) / "t1.nii.gz"
            save_nifti(str(t1_path), amap.astype(np.float32), affine)

            reconst_flow = ReconstCSDFlow()
            reconst_flow._force_overwrite = True

            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                use_msmt=True,
                t1_file=str(t1_path),
                iso=3,
                sh_order_max=4,
                out_dir=out_dir,
                extract_pam_values=True,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            npt.assert_equal(load_nifti_data(gfa_path).shape, volume.shape[:-1])

            for fname in ["wm_mask.nii.gz", "gm_mask.nii.gz", "csf_mask.nii.gz"]:
                mask_path_out = Path(out_dir) / fname
                assert mask_path_out.exists()
                npt.assert_equal(
                    load_nifti_data(str(mask_path_out)).shape, volume.shape[:-1]
                )


def reconst_flow_core(flow, **kwargs):
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_64D")
        volume, affine = load_nifti(data_path)
        mask = np.ones_like(volume[:, :, :, 0])
        mask_path = Path(out_dir) / "tmp_mask.nii.gz"
        save_nifti(mask_path, mask.astype(np.uint8), affine)

        reconst_flow = flow()
        for sh_order in [4, 6, 8]:
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                sh_order_max=sh_order,
                out_dir=out_dir,
                extract_pam_values=True,
                **kwargs,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            gfa_data = load_nifti_data(gfa_path)
            npt.assert_equal(gfa_data.shape, volume.shape[:-1])

            peaks_dir_path = reconst_flow.last_generated_outputs["out_peaks_dir"]
            peaks_dir_data = load_nifti_data(peaks_dir_path)
            npt.assert_equal(peaks_dir_data.shape[-1], 15)
            npt.assert_equal(peaks_dir_data.shape[:-1], volume.shape[:-1])

            peaks_idx_path = reconst_flow.last_generated_outputs["out_peaks_indices"]
            peaks_idx_data = load_nifti_data(peaks_idx_path)
            npt.assert_equal(peaks_idx_data.shape[-1], 5)
            npt.assert_equal(peaks_idx_data.shape[:-1], volume.shape[:-1])

            peaks_vals_path = reconst_flow.last_generated_outputs["out_peaks_values"]
            peaks_vals_data = load_nifti_data(peaks_vals_path)
            npt.assert_equal(peaks_vals_data.shape[-1], 5)
            npt.assert_equal(peaks_vals_data.shape[:-1], volume.shape[:-1])

            shm_path = reconst_flow.last_generated_outputs["out_shm"]
            shm_data = load_nifti_data(shm_path)
            # Test that the number of coefficients is what you would expect
            # given the order of the sh basis:
            npt.assert_equal(
                shm_data.shape[-1], sph_harm_ind_list(sh_order)[0].shape[0]
            )
            npt.assert_equal(shm_data.shape[:-1], volume.shape[:-1])

            pam = load_pam(reconst_flow.last_generated_outputs["out_pam"])
            npt.assert_allclose(
                pam.peak_dirs.reshape(peaks_dir_data.shape), peaks_dir_data
            )
            npt.assert_allclose(pam.peak_values, peaks_vals_data)
            npt.assert_allclose(pam.peak_indices, peaks_idx_data)
            npt.assert_allclose(pam.shm_coeff, shm_data)
            npt.assert_allclose(pam.gfa, gfa_data)

            bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)
            bvals[0] = 5.0
            bvecs = generate_bvecs(len(bvals))

            tmp_bval_path = Path(out_dir) / "tmp.bval"
            tmp_bvec_path = Path(out_dir) / "tmp.bvec"
            np.savetxt(tmp_bval_path, bvals)
            np.savetxt(tmp_bvec_path, bvecs.T)
            reconst_flow._force_overwrite = True

            if flow.get_short_name() == "csd":
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf=[15, 5, 5],
                    **kwargs,
                )
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf="15, 5, 5",
                    **kwargs,
                )
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf=None,
                    **kwargs,
                )
                reconst_flow2 = flow()
                reconst_flow2._force_overwrite = True
                reconst_flow2.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    frf=None,
                    roi_center=[5, 5, 5],
                    **kwargs,
                )
            else:
                with npt.assert_raises(BaseException):
                    assert_warns(
                        UserWarning,
                        reconst_flow.run,
                        data_path,
                        tmp_bval_path,
                        tmp_bvec_path,
                        mask_path,
                        out_dir=out_dir,
                        extract_pam_values=True,
                        **kwargs,
                    )

            # test parallel implementation
            # Avoid SDT for now, as it is quite slow, something to introspect
            if flow.get_short_name() != "sdt":
                reconst_flow = flow()
                reconst_flow._force_overwrite = True
                reconst_flow.run(
                    data_path,
                    bval_path,
                    bvec_path,
                    mask_path,
                    out_dir=out_dir,
                    parallel=True,
                    num_processes=2,
                    **kwargs,
                )


@needs_cvxpy
def test_reconst_csd_msmt_parallel():
    """Test MSMT-CSD with parallel processing."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        warnings.filterwarnings(
            "ignore",
            message="Solution may be inaccurate.*",
            category=UserWarning,
        )
        with TemporaryDirectory() as out_dir:
            data_path, bval_path, bvec_path = get_fnames(name="small_64D")
            volume, affine = load_nifti(data_path)
            mask = np.ones_like(volume[:, :, :, 0])
            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            save_nifti(mask_path, mask.astype(np.uint8), affine)

            reconst_flow = ReconstCSDFlow()
            reconst_flow._force_overwrite = True

            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                use_msmt=True,
                iso=3,
                sh_order_max=4,
                out_dir=out_dir,
                extract_pam_values=True,
                parallel=True,
                num_processes=2,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            npt.assert_equal(load_nifti_data(gfa_path).shape, volume.shape[:-1])


def test_reconst_csd_msmt_invalid_iso():
    """Test that iso < 3 raises an error."""
    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_64D")
        volume, affine = load_nifti(data_path)
        mask = np.ones_like(volume[:, :, :, 0])
        mask_path = Path(out_dir) / "tmp_mask.nii.gz"
        save_nifti(mask_path, mask.astype(np.uint8), affine)

        reconst_flow = ReconstCSDFlow()
        with npt.assert_raises(SystemExit):
            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                use_msmt=True,
                iso=2,
                sh_order_max=4,
                out_dir=out_dir,
            )


@needs_cvxpy
def test_reconst_csd_msmt_with_tissue_masks():
    """Test MSMT-CSD with pre-provided tissue masks (no HMRF classification).

    Tissue masks are derived from the DWI anisotropic power map via HMRF so
    that the tissue labels are consistent with the DWI signal properties,
    avoiding degenerate response functions.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        warnings.filterwarnings(
            "ignore",
            message="Solution may be inaccurate.*",
            category=UserWarning,
        )
        with TemporaryDirectory() as out_dir:
            data_path, bval_path, bvec_path = get_fnames(name="small_64D")
            volume, affine = load_nifti(data_path)
            bvals, bvecs = read_bvals_bvecs(bval_path, bvec_path)
            shape = volume.shape[:3]

            # Derive tissue masks from the DWI-based anisotropic power map so
            # they are consistent with the actual diffusion signal.
            from dipy.core.gradients import gradient_table
            from dipy.core.sphere import HemiSphere
            from dipy.reconst.shm import (
                anisotropic_power,
                normalize_data,
                smooth_pinv,
                sph_harm_lookup,
            )
            from dipy.segment.tissue import TissueClassifierHMRF

            gtab = gradient_table(bvals, bvecs=bvecs)
            dwi_mask_b = ~gtab.b0s_mask
            normed = normalize_data(volume, gtab.b0s_mask)[..., dwi_mask_b]
            signal_pts = HemiSphere(xyz=gtab.bvecs[dwi_mask_b])
            Ba, m, n = sph_harm_lookup.get("descoteaux07")(
                4, signal_pts.theta, signal_pts.phi
            )
            invB = smooth_pinv(Ba, np.sqrt(0.0) * (-n * (n + 1)))
            amap = anisotropic_power(
                np.dot(normed, invB.T), norm_factor=0.00001, power=2, non_negative=True
            )
            hmrf = TissueClassifierHMRF()
            _, seg, _ = hmrf.classify(amap, 3, 0.1)
            wm_mask = np.where(seg == 3, 1, 0).astype(np.uint8)
            gm_mask = np.where(seg == 2, 1, 0).astype(np.uint8)
            csf_mask = np.where(seg == 1, 1, 0).astype(np.uint8)

            mask_path = Path(out_dir) / "tmp_mask.nii.gz"
            wm_path = Path(out_dir) / "wm_in.nii.gz"
            gm_path = Path(out_dir) / "gm_in.nii.gz"
            csf_path = Path(out_dir) / "csf_in.nii.gz"
            save_nifti(mask_path, np.ones(shape, dtype=np.uint8), affine)
            save_nifti(wm_path, wm_mask, affine)
            save_nifti(gm_path, gm_mask, affine)
            save_nifti(csf_path, csf_mask, affine)

            reconst_flow = ReconstCSDFlow()
            reconst_flow._force_overwrite = True

            reconst_flow.run(
                data_path,
                bval_path,
                bvec_path,
                mask_path,
                use_msmt=True,
                wm_file=str(wm_path),
                gm_file=str(gm_path),
                csf_file=str(csf_path),
                iso=3,
                sh_order_max=4,
                out_dir=out_dir,
                extract_pam_values=True,
            )

            gfa_path = reconst_flow.last_generated_outputs["out_gfa"]
            npt.assert_equal(load_nifti_data(gfa_path).shape, volume.shape[:-1])



import nibabel as nib
import numpy as np
from numpy.testing import assert_, assert_almost_equal, assert_equal, assert_raises

from dipy.align.reslice import reslice
from dipy.data import get_fnames
from dipy.denoise.noise_estimate import estimate_sigma
from dipy.io.image import load_nifti


def test_resample():
    fimg, _, _ = get_fnames(name="small_25")
    data, affine, zooms = load_nifti(fimg, return_voxsize=True)

    # test that new zooms are correctly from the affine (check with 3D volume)
    new_zooms = (1, 1.2, 2.1)
    data2, affine2 = reslice(
        data[..., 0], affine, zooms, new_zooms, order=1, mode="constant"
    )
    img2 = nib.Nifti1Image(data2, affine2)
    new_zooms_confirmed = img2.header.get_zooms()[:3]
    assert_almost_equal(new_zooms, new_zooms_confirmed)

    # test that shape changes correctly for the first 3 dimensions (check 4D)
    new_zooms = (1, 1, 1.0)
    data2, affine2 = reslice(data, affine, zooms, new_zooms, order=0, mode="reflect")
    assert_equal(2 * np.array(data.shape[:3]), data2.shape[:3])
    assert_equal(data2.shape[-1], data.shape[-1])

    # same with different interpolation order
    new_zooms = (1, 1, 1.0)
    data3, affine2 = reslice(data, affine, zooms, new_zooms, order=5, mode="reflect")
    assert_equal(2 * np.array(data.shape[:3]), data3.shape[:3])
    assert_equal(data3.shape[-1], data.shape[-1])

    # test that the sigma will be reduced with interpolation
    sigmas = estimate_sigma(data)
    sigmas2 = estimate_sigma(data2)
    sigmas3 = estimate_sigma(data3)

    assert_(np.all(sigmas > sigmas2))
    assert_(np.all(sigmas2 > sigmas3))

    # check that 4D resampling matches 3D resampling
    data2, affine2 = reslice(data, affine, zooms, new_zooms)
    for i in range(data.shape[-1]):
        _data, _affine = reslice(data[..., i], affine, zooms, new_zooms)
        assert_almost_equal(data2[..., i], _data)
        assert_almost_equal(affine2, _affine)

    # check use of multiprocessing pool of specified size
    data3, affine3 = reslice(data, affine, zooms, new_zooms, num_processes=4)
    assert_almost_equal(data2, data3)
    assert_almost_equal(affine2, affine3)

    # check use of multiprocessing pool of autoconfigured size
    data3, affine3 = reslice(data, affine, zooms, new_zooms, num_processes=-1)
    assert_almost_equal(data2, data3)
    assert_almost_equal(affine2, affine3)

    # test invalid values of num_threads
    assert_raises(ValueError, reslice, data, affine, zooms, new_zooms, num_processes=0)

    # test invalid volume dimension
    assert_raises(
        ValueError, reslice, np.zeros((4, 4, 4, 4, 1)), affine, zooms, new_zooms
    )



"""Utilities for testing."""

from functools import partial
import operator
from pathlib import Path
import warnings

import numpy as np
import numpy.testing as npt
from numpy.testing import assert_array_equal

# set path to example data
IO_DATA_PATH = Path(__file__).resolve().parent / ".." / "io" / "tests" / "data"


def assert_operator(value1, value2, msg="", op=operator.eq):
    """Check Boolean statement."""
    try:
        if op == operator.is_:
            value1 = bool(value1)
        assert op(value1, value2)
    except AssertionError as e:
        raise AssertionError(msg.format(str(value2), str(value1))) from e


assert_greater_equal = partial(assert_operator, op=operator.ge, msg="{0} >= {1}")
assert_greater = partial(assert_operator, op=operator.gt, msg="{0} > {1}")
assert_less_equal = partial(assert_operator, op=operator.le, msg="{0} =< {1}")
assert_less = partial(assert_operator, op=operator.lt, msg="{0} < {1}")
assert_true = partial(
    assert_operator, value2=True, op=operator.is_, msg="False is not true"
)
assert_false = partial(
    assert_operator, value2=False, op=operator.is_, msg="True is not false"
)
assert_not_equal = partial(assert_operator, op=operator.ne)


def assert_arrays_equal(arrays1, arrays2):
    for arr1, arr2 in zip(arrays1, arrays2):
        assert_array_equal(arr1, arr2)


def assert_percent_almost_equal(a, b, decimal=7, percent=0.99):
    a = np.asarray(a)
    b = np.asarray(b)
    tol = 1.5 * 10 ** (-decimal)

    diff = np.abs(a - b)
    ok = diff <= tol

    fraction = np.mean(ok)

    if fraction < percent:
        raise AssertionError(
            f"Only {fraction*100:.2f}% of elements match within {decimal} decimals; "
            f"required {percent*100:.2f}%"
        )


class clear_and_catch_warnings(warnings.catch_warnings):
    """Context manager that resets warning registry for catching warnings.

    Warnings can be slippery, because, whenever a warning is triggered, Python
    adds a ``__warningregistry__`` member to the *calling* module.  This makes
    it impossible to retrigger the warning in this module, whatever you put in
    the warnings filters.  This context manager accepts a sequence of `modules`
    as a keyword argument to its constructor and:
    * stores and removes any ``__warningregistry__`` entries in given `modules`
      on entry;
    * resets ``__warningregistry__`` to its previous state on exit.
    This makes it possible to trigger any warning afresh inside the context
    manager without disturbing the state of warnings outside.
    For compatibility with Python 3.0, please consider all arguments to be
    keyword-only.

    Parameters
    ----------
    record : bool, optional
        Specifies whether warnings should be captured by a custom
        implementation of ``warnings.showwarning()`` and be appended to a list
        returned by the context manager. Otherwise None is returned by the
        context manager. The objects appended to the list are arguments whose
        attributes mirror the arguments to ``showwarning()``.
        NOTE: nibabel difference from numpy: default is True
    modules : sequence, optional
        Sequence of modules for which to reset warnings registry on entry and
        restore on exit

    Examples
    --------
    >>> import warnings
    >>> import numpy as np
    >>> with clear_and_catch_warnings(modules=[np.lib.scimath]):
    ...     warnings.simplefilter('always')
    ...     # do something that raises a warning in np.lib.scimath
    ...     _ = np.arccos(90)

    Notes
    -----
    this class is copied (with minor modifications) from Nibabel.
    https://github.com/nipy/nibabel. See COPYING file distributed along with
    the Nibabel package for the copyright and license terms.

    """

    class_modules = ()

    def __init__(self, record=True, modules=()):
        self.modules = set(modules).union(self.class_modules)
        self._warnreg_copies = {}
        super(clear_and_catch_warnings, self).__init__(record=record)

    def __enter__(self):
        for mod in self.modules:
            if hasattr(mod, "__warningregistry__"):
                mod_reg = mod.__warningregistry__
                self._warnreg_copies[mod] = mod_reg.copy()
                mod_reg.clear()
        return super(clear_and_catch_warnings, self).__enter__()

    def __exit__(self, *exc_info):
        super(clear_and_catch_warnings, self).__exit__(*exc_info)
        for mod in self.modules:
            if hasattr(mod, "__warningregistry__"):
                mod.__warningregistry__.clear()
            if mod in self._warnreg_copies:
                mod.__warningregistry__.update(self._warnreg_copies[mod])


def check_for_warnings(warn_printed, w_msg):
    selected_w = [w for w in warn_printed if issubclass(w.category, UserWarning)]
    assert len(selected_w) >= 1
    msg = [str(m.message) for m in selected_w]
    npt.assert_equal(w_msg in msg, True)


def assert_warns(warning_class, *args, **kwargs):
    """Assert that a warning is raised.

    This function can be used as a context manager or as a function wrapper.
    It is a replacement for the deprecated numpy.testing.assert_warns.

    Parameters
    ----------
    warning_class : class
        The class of warning to check for.
    *args : positional arguments
        If provided, the first argument should be a callable, and the
        remaining arguments are passed to that callable.
    **kwargs : keyword arguments
        Keyword arguments passed to the callable.

    Returns
    -------
    ctx : WarningsChecker or result
        If used as a context manager, returns a WarningsChecker instance.
        If used as a function wrapper, returns the result of the callable.

    Examples
    --------
    As a context manager:
    >>> with assert_warns(UserWarning):
    ...     warnings.warn("test", UserWarning)

    As a function wrapper:
    >>> def f():
    ...     warnings.warn("test", UserWarning)
    ...     return 42
    >>> result = assert_warns(UserWarning, f)
    >>> result
    42

    Notes
    -----
    This replaces the deprecated numpy.testing.assert_warns function.
    It uses warnings.catch_warnings internally for better compatibility
    with NumPy 2.x.
    """
    if not args:
        # Used as a context manager
        return _WarningsChecker(warning_class)
    else:
        # Used as a function wrapper
        func = args[0]
        func_args = args[1:]
        with _WarningsChecker(warning_class):
            return func(*func_args, **kwargs)


class _WarningsChecker:
    """Context manager to check that a warning is raised."""

    def __init__(self, warning_class):
        self.warning_class = warning_class
        self.caught_warnings = []

    def __enter__(self):
        self.catch_warnings = warnings.catch_warnings(record=True)
        self.warning_list = self.catch_warnings.__enter__()
        warnings.simplefilter("always")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.catch_warnings.__exit__(exc_type, exc_val, exc_tb)

        # Check if the expected warning was raised
        if exc_type is None:  # No exception was raised
            matching_warnings = [
                w
                for w in self.warning_list
                if issubclass(w.category, self.warning_class)
            ]
            if not matching_warnings:
                raised_warnings = [w.category.__name__ for w in self.warning_list]
                raise AssertionError(
                    f"No warning of class {self.warning_class.__name__} "
                    f"was raised. Warnings raised: {raised_warnings}"
                )
        return False



import warnings

import numpy as np
import numpy.testing as npt
import pytest

from dipy.testing import assert_true, assert_warns
from dipy.testing.decorators import set_random_number_generator
from dipy.tracking import metrics
from dipy.tracking._utils import _to_voxel_coordinates
from dipy.tracking.streamline import transform_streamlines
from dipy.tracking.utils import (
    _min_at,
    clip_streamlines_to_target,
    connectivity_matrix,
    density_map,
    length,
    max_angle_from_curvature,
    min_radius_curvature_from_angle,
    ndbincount,
    near_roi,
    path_length,
    random_seeds_from_mask,
    reduce_labels,
    reduce_rois,
    seeds_directions_pairs,
    seeds_from_mask,
    target,
    target_line_based,
    unique_rows,
)
from dipy.tracking.vox2track import streamline_mapping


def make_streamlines(return_seeds=False):
    streamlines = [
        np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2], [5, 10, 12]], "float"),
        np.array([[1, 2, 3], [3, 2, 0], [5, 20, 33], [40, 80, 120]], "float"),
    ]
    seeds = [np.array([0.0, 0.0, 0.0], "float"), np.array([1.0, 2.0, 3.0], "float")]
    if return_seeds:
        return streamlines, seeds
    else:
        return streamlines


def test_density_map():
    # One streamline diagonal in volume
    streamlines = [np.array([np.arange(10)] * 3).T]
    shape = (10, 10, 10)
    x = np.arange(10)
    expected = np.zeros(shape)
    expected[x, x, x] = 1.0
    dm = density_map(streamlines, np.eye(4), shape)
    npt.assert_array_equal(dm, expected)

    # add streamline, make voxel_size smaller. Each streamline should only be
    # counted once, even if multiple points lie in a voxel
    streamlines.append(np.ones((5, 3)))
    shape = (5, 5, 5)
    x = np.arange(5)
    expected = np.zeros(shape)
    expected[x, x, x] = 1.0
    expected[0, 0, 0] += 1
    affine = np.eye(4) * 2
    affine[:3, 3] = 0.05
    dm = density_map(streamlines, affine, shape)
    npt.assert_array_equal(dm, expected)
    # should work with a generator
    dm = density_map(iter(streamlines), affine, shape)
    npt.assert_array_equal(dm, expected)

    # Test passing affine
    affine = np.diag([2, 2, 2, 1.0])
    affine[:3, 3] = 1.0
    dm = density_map(streamlines, affine, shape)
    npt.assert_array_equal(dm, expected)

    # Shift the image by 2 voxels, ie 4mm
    affine[:3, 3] -= 4.0
    expected_old = expected
    new_shape = [i + 2 for i in shape]
    expected = np.zeros(new_shape)
    expected[2:, 2:, 2:] = expected_old
    dm = density_map(streamlines, affine, new_shape)
    npt.assert_array_equal(dm, expected)


def test_to_voxel_coordinates_precision():
    # To simplify tests, use an identity affine. This would be the result of
    # a call to _mapping_to_voxel with another identity affine.
    transfo = np.array([[1.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 1.0]])

    # Offset is computed by _mapping_to_voxel. With a 1x1x1 dataset
    # having no translation, the offset is half the voxel size, i.e. 0.5.
    offset = np.array([0.5, 0.5, 0.5])

    # Without the added tolerance in _to_voxel_coordinates, this streamline
    # should raise an Error in the call to _to_voxel_coordinates.
    failing_strl = [
        np.array([[-0.5000001, 0.0, 0.0], [0.0, 1.0, 0.0]], dtype=np.float32)
    ]

    indices = _to_voxel_coordinates(failing_strl, transfo, offset)

    expected_indices = np.array([[[0, 0, 0], [0, 1, 0]]])
    npt.assert_array_equal(indices, expected_indices)


def test_connectivity_matrix():
    label_volume = np.array([[[3, 0, 0], [0, 0, 5], [0, 0, 4]]])
    streamlines = [
        np.array([[0, 0, 0], [0, 1, 2], [0, 2, 2]], "float"),
        np.array([[0, 0, 0], [0, 1, 1], [0, 2, 2]], "float"),
        np.array([[0, 2, 2], [0, 1, 1], [0, 0, 0]], "float"),
    ]
    expected = np.zeros((6, 6), "int")
    expected[3, 4] = 2
    expected[4, 3] = 1
    # Check basic Case
    matrix = connectivity_matrix(streamlines, np.eye(4), label_volume, symmetric=False)
    npt.assert_array_equal(matrix, expected)
    # Test mapping
    matrix, mapping = connectivity_matrix(
        streamlines, np.eye(4), label_volume, symmetric=False, return_mapping=True
    )
    npt.assert_array_equal(matrix, expected)
    npt.assert_equal(mapping[3, 4], [0, 1])
    npt.assert_equal(mapping[4, 3], [2])
    npt.assert_equal(mapping.get((0, 0)), None)
    # Test mapping and symmetric
    matrix, mapping = connectivity_matrix(
        streamlines, np.eye(4), label_volume, symmetric=True, return_mapping=True
    )
    npt.assert_equal(mapping[3, 4], [0, 1, 2])
    # When symmetric only (3,4) is a key, not (4, 3)
    npt.assert_equal(mapping.get((4, 3)), None)
    # expected output matrix is symmetric version of expected
    expected = expected + expected.T
    npt.assert_array_equal(matrix, expected)
    # Test mapping_as_streamlines, mapping dict has lists of streamlines
    matrix, mapping = connectivity_matrix(
        streamlines,
        np.eye(4),
        label_volume,
        symmetric=False,
        return_mapping=True,
        mapping_as_streamlines=True,
    )
    assert_true(mapping[3, 4][0] is streamlines[0])
    assert_true(mapping[3, 4][1] is streamlines[1])
    assert_true(mapping[4, 3][0] is streamlines[2])

    # Test Inclusive streamline analysis
    # Check basic Case (inclusive)
    expected = np.zeros((6, 6), "int")
    expected[3, 4] = 2
    expected[4, 3] = 1
    expected[3, 5] = 1
    expected[5, 4] = 1
    expected[0, 3:5] = 1
    expected[3:5, 0] = 1

    matrix = connectivity_matrix(
        streamlines, np.eye(4), label_volume, symmetric=False, inclusive=True
    )
    npt.assert_array_equal(matrix, expected)

    # Test mapping
    matrix, mapping = connectivity_matrix(
        streamlines,
        np.eye(4),
        label_volume,
        inclusive=True,
        symmetric=False,
        return_mapping=True,
    )
    npt.assert_array_equal(matrix, expected)
    npt.assert_equal(mapping[3, 4], [0, 1])
    npt.assert_equal(mapping[4, 3], [2])
    npt.assert_equal(mapping[0, 4], [1])
    npt.assert_equal(mapping[3, 5], [0])
    npt.assert_equal(mapping.get((0, 0)), None)

    # Test mapping and symmetric
    matrix, mapping = connectivity_matrix(
        streamlines,
        np.eye(4),
        label_volume,
        inclusive=True,
        symmetric=True,
        return_mapping=True,
    )
    npt.assert_equal(mapping[3, 4], [0, 1, 2])
    npt.assert_equal(mapping[0, 3], [1, 2])
    npt.assert_equal(mapping[0, 4], [1, 2])
    npt.assert_equal(mapping[3, 5], [0])
    npt.assert_equal(mapping[4, 5], [0])

    # When symmetric only (3,4) is a key, not (4,3)
    npt.assert_equal(mapping.get((4, 3)), None)

    # expected output matrix is symmetric version of expected
    expected = expected + expected.T
    npt.assert_array_equal(matrix, expected)
    # Test mapping_as_streamlines, mapping dict has lists of streamlines
    matrix, mapping = connectivity_matrix(
        streamlines,
        np.eye(4),
        label_volume,
        symmetric=False,
        inclusive=True,
        return_mapping=True,
        mapping_as_streamlines=True,
    )
    assert_true(mapping[0, 3][0] is streamlines[2])
    assert_true(mapping[0, 4][0] is streamlines[1])
    assert_true(mapping[3, 0][0] is streamlines[1])
    assert_true(mapping[3, 4][0] is streamlines[0])
    assert_true(mapping[3, 4][1] is streamlines[1])
    assert_true(mapping[3, 5][0] is streamlines[0])
    assert_true(mapping[4, 0][0] is streamlines[2])
    assert_true(mapping[4, 3][0] is streamlines[2])
    assert_true(mapping[5, 4][0] is streamlines[0])

    # Test weighted streamline analysis
    weights = np.array([1.5, 0, 3])

    # Check weighted non-inclusive case
    expected = np.zeros((6, 6))
    expected[3, 4] = 1.5
    expected[4, 3] = 3
    matrix = connectivity_matrix(
        streamlines, np.eye(4), label_volume, symmetric=False, weights=weights
    )
    npt.assert_array_equal(matrix, expected)

    # Check weighted inclusive case
    expected = np.zeros((6, 6))
    expected[3, 4] = 1.5
    expected[4, 3] = 3
    expected[3, 5] = 1.5
    expected[5, 4] = 1.5
    expected[0, 3] = 3
    expected[4, 0] = 3

    matrix = connectivity_matrix(
        streamlines,
        np.eye(4),
        label_volume,
        symmetric=False,
        inclusive=True,
        weights=weights,
    )
    npt.assert_array_equal(matrix, expected)

    # Test passing affine to connectivity_matrix
    affine = np.diag([-1, -1, -1, 1.0])
    streamlines = [-i for i in streamlines]
    matrix = connectivity_matrix(streamlines, affine, label_volume)
    # In the symmetrical case, the matrix should be, well, symmetric:
    npt.assert_equal(matrix[4, 3], matrix[4, 3])


def test_connectivity_matrix_with_generator():
    """Test connectivity_matrix works with generator inputs."""
    label_volume = np.array([[[3, 0, 0], [0, 0, 5], [0, 0, 4]]])
    streamlines = [
        np.array([[0, 0, 0], [0, 1, 2], [0, 2, 2]], "float"),
        np.array([[0, 0, 0], [0, 1, 1], [0, 2, 2]], "float"),
        np.array([[0, 2, 2], [0, 1, 1], [0, 0, 0]], "float"),
    ]

    # Create a generator version of streamlines
    def streamline_generator():
        for sl in streamlines:
            yield sl

    # Test that generator input works
    matrix = connectivity_matrix(
        streamline_generator(), np.eye(4), label_volume, symmetric=False
    )
    expected = np.zeros((6, 6), "int")
    expected[3, 4] = 2
    expected[4, 3] = 1
    npt.assert_array_equal(matrix, expected)

    # Test with discard_stream_size parameter
    matrix = connectivity_matrix(
        streamline_generator(),
        np.eye(4),
        label_volume,
        symmetric=False,
        discard_stream_size=1,
    )
    npt.assert_array_equal(matrix, expected)

    # Test inclusive mode with generator
    expected_inclusive = np.zeros((6, 6), "int")
    expected_inclusive[3, 4] = 2
    expected_inclusive[4, 3] = 1
    expected_inclusive[3, 5] = 1
    expected_inclusive[5, 4] = 1
    expected_inclusive[0, 3:5] = 1
    expected_inclusive[3:5, 0] = 1

    matrix = connectivity_matrix(
        streamline_generator(),
        np.eye(4),
        label_volume,
        symmetric=False,
        inclusive=True,
        discard_stream_size=1,
    )
    npt.assert_array_equal(matrix, expected_inclusive)


def test_ndbincount():
    def check(expected):
        npt.assert_equal(bc[0, 0], expected[0])
        npt.assert_equal(bc[0, 1], expected[1])
        npt.assert_equal(bc[1, 0], expected[2])
        npt.assert_equal(bc[2, 2], expected[3])

    x = np.array([[0, 0], [0, 0], [0, 1], [0, 1], [1, 0], [2, 2]]).T
    expected = [2, 2, 1, 1]
    # count occurrences in x
    bc = ndbincount(x)
    npt.assert_equal(bc.shape, (3, 3))
    check(expected)
    # pass in shape
    bc = ndbincount(x, shape=(4, 5))
    npt.assert_equal(bc.shape, (4, 5))
    check(expected)
    # pass in weights
    weights = np.arange(6.0)
    weights[-1] = 1.23
    expected = [1.0, 5.0, 4.0, 1.23]
    bc = ndbincount(x, weights=weights)
    check(expected)
    # raises an error if shape is too small
    npt.assert_raises(ValueError, ndbincount, x, weights=None, shape=(2, 2))


def test_reduce_labels():
    shape = (4, 5, 6)
    # labels from 100 to 220
    labels = np.arange(100, np.prod(shape) + 100).reshape(shape)
    # new labels form 0 to 120, and lookup maps range(0,120) to range(100, 220)
    new_labels, lookup = reduce_labels(labels)
    npt.assert_array_equal(new_labels, labels - 100)
    npt.assert_array_equal(lookup, labels.ravel())


def test_target():
    streamlines = [
        np.array([[0.0, 0.0, 0.0], [1.0, 0.0, 0.0], [2.0, 0.0, 0.0]]),
        np.array([[0.0, 0.0, 0], [0, 1.0, 1.0], [0, 2.0, 2.0]]),
    ]
    _target(target, streamlines, (0, 0, 0), (1, 0, 0), True)


def test_target_lb():
    streamlines = [
        np.array([[0.0, 1.0, 1.0], [3.0, 1.0, 1.0]]),
        np.array([[0.0, 0.0, 0.0], [2.0, 2.0, 2.0]]),
        np.array([[1.0, 1.0, 1.0]]),
    ]  # Single-point streamline
    _target(target_line_based, streamlines, (1, 1, 1), (2, 1, 1), False)


def _target(target_f, streamlines, voxel_both_true, voxel_one_true, test_bad_points):
    affine = np.eye(4)
    mask = np.zeros((4, 4, 4), dtype=bool)

    # Both pass though
    mask[voxel_both_true] = True
    new = list(target_f(streamlines, affine, mask))
    npt.assert_equal(len(new), 2)
    new = list(target_f(streamlines, affine, mask, include=False))
    npt.assert_equal(len(new), 0)

    # only first
    mask[:] = False
    mask[voxel_one_true] = True
    new = list(target_f(streamlines, affine, mask))
    npt.assert_equal(len(new), 1)
    assert_true(new[0] is streamlines[0])
    new = list(target_f(streamlines, affine, mask, include=False))
    npt.assert_equal(len(new), 1)
    assert_true(new[0] is streamlines[1])

    # Test that bad points raise a value error
    if test_bad_points:
        bad_sl = streamlines + [np.array([[10.0, 10.0, 10.0]])]
        new = target_f(bad_sl, affine, mask)
        npt.assert_raises(ValueError, list, new)
        bad_sl = streamlines + [-np.array([[10.0, 10.0, 10.0]])]
        new = target_f(bad_sl, affine, mask)
        npt.assert_raises(ValueError, list, new)

    # Test smaller voxels
    affine = np.array([[0.3, 0, 0, 0], [0, 0.2, 0, 0], [0, 0, 0.4, 0], [0, 0, 0, 1]])
    streamlines = transform_streamlines(streamlines, affine)
    new = list(target_f(streamlines, affine, mask))
    npt.assert_equal(len(new), 1)
    assert_true(new[0] is streamlines[0])
    new = list(target_f(streamlines, affine, mask, include=False))
    npt.assert_equal(len(new), 1)
    assert_true(new[0] is streamlines[1])

    # Test that changing mask or affine does not break target/target_line_based
    include = target_f(streamlines, affine, mask)
    exclude = target_f(streamlines, affine, mask, include=False)
    affine[:] = np.eye(4)
    mask[:] = False
    include = list(include)
    exclude = list(exclude)
    npt.assert_equal(len(include), 1)
    assert_true(include[0] is streamlines[0])
    npt.assert_equal(len(exclude), 1)
    assert_true(exclude[0] is streamlines[1])


def test_target_line_based_out_of_bounds():
    test_cases = [
        (np.array([[-10, 0, 0], [0, 0, 0]]), 0),
        (np.array([[-10, 0, 0], [1, -10, 0]]), 0),
        (np.array([[0, 0, 0], [10, 10, 10]]), 0),
        (np.array([[-2, -0.6, -0.6], [2, -0.6, -0.6]]), 0),
        (np.array([[-10000, 0, 0], [0, 0, 0]]), 0),
        (np.array([[-10, 0, 0], [10, 0, 0]]), 1),
        (np.array([[1, -10, 0], [1, 10, 0]]), 1),
    ]
    for streamline, expected_matched in test_cases:
        mask = np.zeros((2, 1, 1), dtype=np.uint8)
        mask[1, 0, 0] = 1
        matched = list(target_line_based([streamline], np.eye(4), mask))
        assert len(matched) == expected_matched


def test_clip_to_target():
    aff = np.eye(4)

    sl1 = np.zeros([5, 3])
    sl1[:, 0] = np.arange(5)
    sl1[:, 2] = 2
    sl2 = sl1.copy()
    sl2[:, 1] = 1
    sl3 = sl1.copy()
    sl3[:, 1] = 2
    sl4 = sl1.copy()
    sl4[:, 1] = 40

    streamlines = [sl1] + [sl2] + [sl3]
    streamlines_outside = [sl1] + [sl2] + [sl3] + [sl4]

    # clipping to length 3 if streamline touches
    roi1 = np.zeros([20, 20, 20])  # all clipped
    roi1[3:6, 0:5, 0:5] = 1
    new1 = list(clip_streamlines_to_target(streamlines, roi1, affine=aff))
    npt.assert_array_equal([len(i) for i in new1], np.array([3, 3, 3]))

    roi2 = np.zeros([20, 20, 20])  # one clipped
    roi2[3:6, 1:5, 0:5] = 1
    new2 = list(clip_streamlines_to_target(streamlines, roi2, affine=aff))
    npt.assert_array_equal([len(i) for i in new2], np.array([5, 3, 3]))

    roi3 = np.zeros([20, 20, 20])  # two clipped
    roi3[3:6, 2:5, 0:5] = 1
    new3 = list(clip_streamlines_to_target(streamlines, roi3, affine=aff))
    npt.assert_array_equal([len(i) for i in new3], np.array([5, 5, 3]))

    # check if sls pass through roi instead of terminate in it
    roi4 = np.zeros([20, 20, 20])
    roi4[3:4, 0:5, 0:5] = 1
    new4 = list(clip_streamlines_to_target(streamlines, roi4, affine=aff))
    npt.assert_array_equal([len(i) for i in new4], [len(i) for i in new1])

    # check if roi on either side it picks bigger piece of streamline
    roi5 = np.zeros([20, 20, 20])
    roi5[4:5, 0:5, 0:5] = 1
    new5 = list(clip_streamlines_to_target(streamlines, roi5, affine=aff))
    npt.assert_array_equal([len(i) for i in new5], np.array([4, 4, 4]))

    roi5 = np.zeros([20, 20, 20])
    roi5[1:2, 0:5, 0:5] = 1
    new5 = list(clip_streamlines_to_target(streamlines, roi5, affine=aff))
    npt.assert_array_equal([len(i) for i in new5], np.array([4, 4, 4]))

    # check value error for streamline outside of target mask dimensions
    bad = clip_streamlines_to_target(streamlines_outside, roi1, affine=aff)
    npt.assert_raises(ValueError, list, bad)

    # Test smaller voxels
    affine = np.array([[0.3, 0, 0, 0], [0, 0.2, 0, 0], [0, 0, 0.4, 0], [0, 0, 0, 1]])
    new1 = list(clip_streamlines_to_target(streamlines, roi1, affine=affine))
    npt.assert_array_equal([len(i) for i in new1], np.array([5, 5, 5]))
    # Test with an affine that includes a translation
    affine = np.array([[1, 0, 0, -1.0], [0, 1, 0, 0.5], [0, 0, 1, 0.5], [0, 0, 0, 1]])
    new1 = list(clip_streamlines_to_target(streamlines, roi1, affine=affine))
    npt.assert_array_equal(new1, np.array([]))


def test_near_roi():
    streamlines = [
        np.array([[0.0, 0.0, 0.9], [1.9, 0.0, 0.0], [3, 2.0, 2.0]]),
        np.array([[0.1, 0.0, 0], [0, 1.0, 1.0], [0, 2.0, 2.0]]),
        np.array([[2, 2, 2], [3, 3, 3]]),
    ]

    mask = np.zeros((4, 4, 4), dtype=bool)
    mask[0, 0, 0] = True
    mask[1, 0, 0] = True

    npt.assert_array_equal(
        near_roi(streamlines, np.eye(4), mask, tol=1), np.array([True, True, False])
    )
    npt.assert_array_equal(
        near_roi(streamlines, np.eye(4), mask), np.array([False, True, False])
    )

    # test for handling of various forms of null streamlines
    # including a streamline from previous test because near_roi / tol
    # can't handle completely empty streamline collections
    streamlines_null = [
        np.array([[0.0, 0.0, 0.9], [1.9, 0.0, 0.0], [3, 2.0, 2.0]]),
        np.array([[], [], []]).T,
        np.array([]),
        [],
    ]
    npt.assert_array_equal(
        near_roi(streamlines_null, np.eye(4), mask, tol=1),
        np.array([True, False, False, False]),
    )
    npt.assert_array_equal(
        near_roi(streamlines_null, np.eye(4), mask),
        np.array([False, False, False, False]),
    )

    # If there is an affine, we need to use it:
    affine = np.eye(4)
    affine[:, 3] = [-1, 100, -20, 1]
    # Transform the streamlines:
    x_streamlines = [sl + affine[:3, 3] for sl in streamlines]
    npt.assert_array_equal(
        near_roi(x_streamlines, affine, mask, tol=1), np.array([True, True, False])
    )
    npt.assert_array_equal(
        near_roi(x_streamlines, affine, mask, tol=None), np.array([False, True, False])
    )

    # Test for use of the 'all' mode:
    npt.assert_array_equal(
        near_roi(x_streamlines, affine, mask, tol=None, mode="all"),
        np.array([False, False, False]),
    )

    mask[0, 1, 1] = True
    mask[0, 2, 2] = True
    # Test for use of the 'all' mode, also testing that setting the tolerance
    # to a very small number gets overridden:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        npt.assert_array_equal(
            near_roi(x_streamlines, affine, mask, tol=0.1, mode="all"),
            np.array([False, True, False]),
        )

    mask[2, 2, 2] = True
    mask[3, 3, 3] = True
    npt.assert_array_equal(
        near_roi(x_streamlines, affine, mask, tol=None, mode="all"),
        np.array([False, True, True]),
    )

    # Test for use of endpoints as selection criteria:
    mask = np.zeros((4, 4, 4), dtype=bool)
    mask[0, 1, 1] = True
    mask[3, 2, 2] = True

    npt.assert_array_equal(
        near_roi(streamlines, np.eye(4), mask, tol=0.87, mode="either_end"),
        np.array([True, False, False]),
    )

    npt.assert_array_equal(
        near_roi(streamlines, np.eye(4), mask, tol=0.87, mode="both_end"),
        np.array([False, False, False]),
    )

    mask[0, 0, 0] = True
    mask[0, 2, 2] = True
    npt.assert_array_equal(
        near_roi(streamlines, np.eye(4), mask, mode="both_end"),
        np.array([False, True, False]),
    )

    # Test with a generator input:
    def generate_sl(streamlines):
        for sl in streamlines:
            yield sl

    npt.assert_array_equal(
        near_roi(generate_sl(streamlines), np.eye(4), mask, mode="both_end"),
        np.array([False, True, False]),
    )


def test_streamline_mapping():
    streamlines = [
        np.array([[0, 0, 0], [0, 0, 0], [0, 2, 2]], "float"),
        np.array([[0, 0, 0], [0, 1, 1], [0, 2, 2]], "float"),
        np.array([[0, 2, 2], [0, 1, 1], [0, 0, 0]], "float"),
    ]
    mapping = streamline_mapping(streamlines, affine=np.eye(4))
    expected = {(0, 0, 0): [0, 1, 2], (0, 2, 2): [0, 1, 2], (0, 1, 1): [1, 2]}
    npt.assert_equal(mapping, expected)

    mapping = streamline_mapping(
        streamlines, affine=np.eye(4), mapping_as_streamlines=True
    )
    expected = {k: [streamlines[i] for i in indices] for k, indices in expected.items()}
    npt.assert_equal(mapping, expected)

    # Test passing affine
    affine = np.eye(4)
    affine[:3, 3] = 0.5
    mapping = streamline_mapping(
        streamlines, affine=affine, mapping_as_streamlines=True
    )
    npt.assert_equal(mapping, expected)

    # Make the voxel size smaller
    affine = np.diag([0.5, 0.5, 0.5, 1.0])
    affine[:3, 3] = 0.25
    expected = {tuple(i * 2 for i in key): value for key, value in expected.items()}
    mapping = streamline_mapping(
        streamlines, affine=affine, mapping_as_streamlines=True
    )
    npt.assert_equal(mapping, expected)


@set_random_number_generator()
def test_length(rng):
    # Generate a simulated bundle of fibers:
    n_streamlines = 50
    n_pts = 100
    t = np.linspace(-10, 10, n_pts)

    bundle = []
    for i in np.linspace(3, 5, n_streamlines):
        pts = np.vstack((np.cos(2 * t / np.pi), np.zeros(t.shape) + i, t)).T
        bundle.append(pts)

    start = rng.integers(10, 30, n_streamlines)
    end = rng.integers(60, 100, n_streamlines)

    bundle = [
        10 * streamline[start[i] : end[i]] for (i, streamline) in enumerate(bundle)
    ]

    bundle_lengths = length(bundle)
    for idx, this_length in enumerate(bundle_lengths):
        npt.assert_equal(this_length, metrics.length(bundle[idx]))


@set_random_number_generator()
def test_seeds_from_mask(rng):
    mask = rng.integers(0, 1, size=(10, 10, 10))
    seeds = seeds_from_mask(mask, np.eye(4), density=1)
    npt.assert_equal(mask.sum(), len(seeds))
    npt.assert_array_equal(np.argwhere(mask), seeds)

    mask[:] = False
    mask[3, 3, 3] = True
    seeds = seeds_from_mask(mask, np.eye(4), density=[3, 4, 5])
    npt.assert_equal(len(seeds), 3 * 4 * 5)
    assert_true(np.all((seeds > 2.5) & (seeds < 3.5)))

    mask[4, 4, 4] = True
    seeds = seeds_from_mask(mask, np.eye(4), density=[3, 4, 5])
    npt.assert_equal(len(seeds), 2 * 3 * 4 * 5)
    assert_true(np.all((seeds > 2.5) & (seeds < 4.5)))
    in_333 = ((seeds > 2.5) & (seeds < 3.5)).all(1)
    npt.assert_equal(in_333.sum(), 3 * 4 * 5)
    in_444 = ((seeds > 3.5) & (seeds < 4.5)).all(1)
    npt.assert_equal(in_444.sum(), 3 * 4 * 5)


@set_random_number_generator()
def test_random_seeds_from_mask(rng):
    mask = rng.integers(0, 1, size=(4, 6, 3))
    seeds = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=24, seed_count_per_voxel=True
    )
    npt.assert_equal(mask.sum() * 24, len(seeds))
    seeds = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=0, seed_count_per_voxel=True
    )
    npt.assert_equal(0, len(seeds))

    mask[:] = False
    mask[2, 2, 2] = True
    seeds = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=8, seed_count_per_voxel=True
    )
    npt.assert_equal(mask.sum() * 8, len(seeds))
    assert_true(np.all((seeds > 1.5) & (seeds < 2.5)))

    seeds = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=24, seed_count_per_voxel=False
    )
    npt.assert_equal(24, len(seeds))
    seeds = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=0, seed_count_per_voxel=False
    )
    npt.assert_equal(0, len(seeds))

    mask[:] = False
    mask[2, 2, 2] = True
    seeds = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=100, seed_count_per_voxel=False
    )
    npt.assert_equal(100, len(seeds))
    assert_true(np.all((seeds > 1.5) & (seeds < 2.5)))

    mask = np.zeros((15, 15, 15))
    mask[2:14, 2:14, 2:14] = 1
    seeds_npv_2 = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=2, seed_count_per_voxel=True, random_seed=0
    )[:150]
    seeds_npv_3 = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=3, seed_count_per_voxel=True, random_seed=0
    )[:150]
    assert_true(np.all(seeds_npv_2 == seeds_npv_3))

    seeds_nt_150 = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=150, seed_count_per_voxel=False, random_seed=0
    )[:150]
    seeds_nt_500 = random_seeds_from_mask(
        mask, np.eye(4), seeds_count=500, seed_count_per_voxel=False, random_seed=0
    )[:150]
    assert_true(np.all(seeds_nt_150 == seeds_nt_500))


def test_connectivity_matrix_shape():
    # Labels: z-planes have labels 0,1,2
    labels = np.zeros((3, 3, 3), dtype=int)
    labels[:, :, 1] = 1
    labels[:, :, 2] = 2
    # Streamline set, only moves between first two z-planes.
    streamlines = [
        np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.5], [0.0, 0.0, 1.0]]),
        np.array([[0.0, 1.0, 1.0], [0.0, 1.0, 0.5], [0.0, 1.0, 0.0]]),
    ]
    matrix = connectivity_matrix(streamlines, np.eye(4), labels)
    npt.assert_equal(matrix.shape, (3, 3))


def test_unique_rows():
    """
    Testing the function unique_coords
    """
    arr = np.array([[1, 2, 3], [1, 2, 3], [2, 3, 4], [3, 4, 5]])
    arr_w_unique = np.array([[1, 2, 3], [2, 3, 4], [3, 4, 5]])
    npt.assert_array_equal(unique_rows(arr), arr_w_unique)

    # Should preserve order:
    arr = np.array([[2, 3, 4], [1, 2, 3], [1, 2, 3], [3, 4, 5]])
    arr_w_unique = np.array([[2, 3, 4], [1, 2, 3], [3, 4, 5]])
    npt.assert_array_equal(unique_rows(arr), arr_w_unique)

    # Should work even with longer arrays:
    arr = np.array(
        [[2, 3, 4], [1, 2, 3], [1, 2, 3], [3, 4, 5], [6, 7, 8], [0, 1, 0], [1, 0, 1]]
    )
    arr_w_unique = np.array(
        [[2, 3, 4], [1, 2, 3], [3, 4, 5], [6, 7, 8], [0, 1, 0], [1, 0, 1]]
    )

    npt.assert_array_equal(unique_rows(arr), arr_w_unique)


def test_reduce_rois():
    roi1 = np.zeros((4, 4, 4), dtype=bool)
    roi2 = np.zeros((4, 4, 4), dtype=bool)
    roi1[1, 1, 1] = 1
    roi2[2, 2, 2] = 1
    include_roi, exclude_roi = reduce_rois([roi1, roi2], [True, True])
    npt.assert_equal(include_roi, roi1 + roi2)
    npt.assert_equal(exclude_roi, np.zeros((4, 4, 4)))
    include_roi, exclude_roi = reduce_rois([roi1, roi2], [True, False])
    npt.assert_equal(include_roi, roi1)
    npt.assert_equal(exclude_roi, roi2)
    # Array input:
    include_roi, exclude_roi = reduce_rois(np.array([roi1, roi2]), [True, True])
    npt.assert_equal(include_roi, roi1 + roi2)
    npt.assert_equal(exclude_roi, np.zeros((4, 4, 4)))
    # Int and float input
    roi1 = np.zeros((4, 4, 4), dtype=int)
    roi2 = np.zeros((4, 4, 4), dtype=float)
    assert_warns(UserWarning, reduce_rois, [roi1], [True])
    assert_warns(UserWarning, reduce_rois, [roi2], [True])


@set_random_number_generator()
def test_path_length(rng):
    aoi = np.zeros((20, 20, 20), dtype=bool)
    aoi[0, 0, 0] = 1

    # A few tests for basic usage
    x = np.arange(20)
    streamlines = [np.array([x, x, x]).T]
    pl = path_length(streamlines, np.eye(4), aoi)
    expected = x.copy() * np.sqrt(3)
    # expected[0] = np.inf
    npt.assert_array_almost_equal(pl[x, x, x], expected)

    aoi[19, 19, 19] = 1
    pl = path_length(streamlines, np.eye(4), aoi)
    expected = np.minimum(expected, expected[::-1])
    npt.assert_array_almost_equal(pl[x, x, x], expected)

    aoi[19, 19, 19] = 0
    aoi[1, 1, 1] = 1
    pl = path_length(streamlines, np.eye(4), aoi)
    expected = (x - 1) * np.sqrt(3)
    expected[0] = 0
    npt.assert_array_almost_equal(pl[x, x, x], expected)

    z = np.zeros(x.shape, x.dtype)
    streamlines.append(np.array([x, z, z]).T)
    pl = path_length(streamlines, np.eye(4), aoi)
    npt.assert_array_almost_equal(pl[x, x, x], expected)
    npt.assert_array_almost_equal(pl[x, 0, 0], x)

    # Only streamlines that pass through aoi contribute to path length so if
    # all streamlines are duds, plm will be all inf.
    aoi[:] = 0
    aoi[0, 0, 0] = 1
    streamlines = []
    for _ in range(1000):
        rando = rng.random(size=(100, 3)) * 19 + 0.5
        assert (rando > 0.5).all()
        assert (rando < 19.5).all()
        streamlines.append(rando)
    pl = path_length(streamlines, np.eye(4), aoi)
    npt.assert_array_almost_equal(pl, -1)

    pl = path_length(streamlines, np.eye(4), aoi, fill_value=-12.0)
    npt.assert_array_almost_equal(pl, -12.0)


def test_min_at():
    k = np.array([3, 2, 2, 2, 1, 1, 1])
    values = np.array([10.0, 1, 2, 3, 31, 21, 11])
    i = np.zeros(k.shape, int)
    j = np.zeros(k.shape, int)
    a = np.zeros([1, 1, 4]) + 100.0

    _min_at(a, (i, j, k), values)
    npt.assert_array_equal(a, [[[100, 11, 1, 10]]])


def test_curvature_angle():
    angle = [0.0000001, np.pi / 3, np.pi / 2.01]
    step_size = [0.2, 0.5, 1.5]
    curvature = [2000000.0, 0.5, 1.064829060280437]

    for theta, step, curve in zip(angle, step_size, curvature):
        res_angle = max_angle_from_curvature(curve, step)
        npt.assert_almost_equal(res_angle, theta)

        res_curvature = min_radius_curvature_from_angle(theta, step)
        npt.assert_almost_equal(res_curvature, curve)

    # special case
    with pytest.warns(UserWarning):
        npt.assert_equal(
            min_radius_curvature_from_angle(0, 1),
            min_radius_curvature_from_angle(np.pi / 2, 1),
        )


@set_random_number_generator()
def test_seeds_directions_pairs(rng):
    positions = rng.random(size=(10, 3))
    peaks = rng.random(size=(10, 5, 3))

    # test various max_cross value
    for i in range(-1, 7):
        seeds, directions = seeds_directions_pairs(positions, peaks, max_cross=i)
        npt.assert_equal(seeds.shape, directions.shape)

    # test it raise and error if the input array shapes don't match
    npt.assert_raises(ValueError, seeds_directions_pairs, positions[:, :2], peaks)
    npt.assert_raises(ValueError, seeds_directions_pairs, positions[:5, :], peaks)
    npt.assert_raises(ValueError, seeds_directions_pairs, positions, peaks[:, :, :2])
    npt.assert_raises(ValueError, seeds_directions_pairs, positions, peaks.flatten())



from pathlib import Path
import sys

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.denoise.bias_correction import bias_field_correction
from dipy.io.gradients import read_bvals_bvecs
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.nn.deepn4 import DeepN4
from dipy.nn.evac import EVACPlus
from dipy.utils.logging import logger
from dipy.workflows.workflow import Workflow


class EVACPlusFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "evacplus"

    def run(
        self,
        input_files,
        save_masked=False,
        out_dir="",
        out_mask="brain_mask.nii.gz",
        out_masked="brain_masked.nii.gz",
    ):
        """Extract brain using EVAC+.

        See :footcite:p:`Park2024` for further details about EVAC+.

        Parameters
        ----------
        input_files : string or Path
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        save_masked : bool, optional
            Save mask.
        out_dir : string or Path, optional
            Output directory.
        out_mask : string, optional
            Name of the mask volume to be saved.
        out_masked : string, optional
            Name of the masked volume to be saved.

        References
        ----------
        .. footbibliography::
        """
        io_it = self.get_io_iterator()
        empty_flag = True

        for fpath, mask_out_path, masked_out_path in io_it:
            logger.info(f"Applying evac+ brain extraction on {fpath}")

            data, affine, img, voxsize = load_nifti(
                fpath, return_img=True, return_voxsize=True
            )
            evac = EVACPlus()
            mask_volume = evac.predict(data, affine)
            masked_volume = mask_volume * data

            save_nifti(mask_out_path, mask_volume.astype(np.float64), affine)

            logger.info(f"Mask saved as {mask_out_path}")

            if save_masked:
                save_nifti(masked_out_path, masked_volume, affine, hdr=img.header)

                logger.info(f"Masked volume saved as {masked_out_path}")
            empty_flag = False
        if empty_flag:
            raise ValueError(
                "All output paths exists."
                " If you want to overwrite "
                "please use the --force option."
            )

        return io_it


class BiasFieldCorrectionFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "bias_field_correction"

    def run(
        self,
        input_files,
        bval=None,
        bvec=None,
        mask=None,
        method="auto",
        threshold=0.5,
        use_cuda=False,
        verbose=False,
        order=3,
        n_control_points=8,
        pyramid_levels="4,2,1",
        n_iter=4,
        lambda_reg=1e-3,
        robust=True,
        gradient_weighting=True,
        zero_background=False,
        out_dir="",
        out_corrected="biasfield_corrected.nii.gz",
        out_bias_field="bias_field.nii.gz",
    ):
        """Correct bias field.

        Parameters
        ----------
        input_files : string or Path
            Path to the input volumes. This path may contain wildcards to
            process multiple inputs at once.
        bval : string or Path, optional
            Path to the b-value file.
        bvec : string or Path, optional
            Path to the b-vector file.
        mask : string or Path, optional
            Path to the brain mask file. If not provided, a mask will be estimated
            from the input volume using Otsu's method for 'poly' and 'bspline' methods.
        method : string, optional
            Bias field correction method. Choose from:
                - 'n4': DeepN4 bias field correction.
                  See :footcite:p:`Kanakaraj2024` for more details.
                - 'poly': Legendre polynomial regression bias correction.
                - 'bspline': Cubic B-spline regression bias correction.
                - 'auto': Run both poly and bspline, return the one with
                  lower Coefficient of Variation within the brain mask.

            'n4' method is recommended for T1-weighted images. 'poly' and 'bspline'
            methods are recommended for diffusion-weighted images.
        threshold : float, optional
            Threshold for cleaning the final correction field in DeepN4 method.
        use_cuda : bool, optional
            Use CUDA for DeepN4 bias field correction.
        verbose : bool, optional
            Print verbose output.
        order : int, optional
            Maximum Legendre polynomial degree (used with method='poly').
        n_control_points : int, optional
            Control grid size per axis (used with method='bspline').
        pyramid_levels : string, optional
            Comma-separated downsampling factors for coarse-to-fine pyramid,
            e.g. '4,2,1' (used with method='poly' or 'bspline').
        n_iter : int, optional
            Reweighting iterations per pyramid level (poly/bspline methods).
        lambda_reg : float, optional
            Ridge regularization strength (poly/bspline methods).
        robust : bool, optional
            Apply Tukey biweight robust reweighting (poly/bspline methods).
        gradient_weighting : bool, optional
            Apply gradient-based edge suppression (poly/bspline methods).
        zero_background : bool, optional
            If True, set the saved bias field to 1.0 outside the brain mask,
            suppressing extrapolation artifacts in the background
            (poly/bspline methods).
        out_dir : string or Path, optional
            Output directory.
        out_corrected : string, optional
            Name of the corrected volume to be saved.
        out_bias_field : string, optional
            Name of the bias field volume to be saved (poly/bspline methods).

        References
        ----------
        .. footbibliography::

        """
        io_it = self.get_io_iterator()

        if method.lower() not in ["n4", "poly", "bspline", "auto"]:
            logger.error(
                "Unknown bias field correction method. "
                "Choose from 'n4', 'poly', 'bspline', 'auto'."
            )
            sys.exit(1)

        prefix = "t1" if method.lower() == "n4" else "dwi"
        for i, name in enumerate(self.flat_outputs):
            if str(name).endswith("biasfield_corrected.nii.gz"):
                self.flat_outputs[i] = Path(name).parent / Path(
                    "biasfield_corrected.nii.gz"
                ).with_name(f"{prefix}_biasfield_corrected.nii.gz")

        if io_it:
            self.update_flat_outputs(self.flat_outputs, io_it)
        else:
            if isinstance(self.last_generated_outputs, dict):
                self.last_generated_outputs = dict(
                    zip(self.last_generated_outputs.keys(), self.flat_outputs)
                )
            else:
                self.last_generated_outputs = self.flat_outputs
        for fpath, corrected_out_path, obf in io_it:
            logger.info(f"Applying bias field correction on {fpath}")

            data, affine, img, voxsize = load_nifti(
                fpath, return_img=True, return_voxsize=True
            )

            corrected_data = None
            if method.lower() == "n4":
                deepn4_model = DeepN4(verbose=verbose, use_cuda=use_cuda)
                deepn4_model.fetch_default_weights()
                corrected_data = deepn4_model.predict(data, affine, threshold=threshold)
            elif method.lower() in ["poly", "bspline", "auto"]:
                bvals, bvecs = read_bvals_bvecs(bval, bvec)
                gtab = gradient_table(bvals, bvecs=bvecs)
                levels = tuple(int(x.strip()) for x in pyramid_levels.split(","))
                n_ctrl = (int(n_control_points),) * 3
                mask_arr = load_nifti_data(mask) if mask is not None else None
                corrected_data, bias = bias_field_correction(
                    data,
                    gtab,
                    mask=mask_arr,
                    method=method.lower(),
                    order=int(order),
                    n_control_points=n_ctrl,
                    pyramid_levels=levels,
                    n_iter=int(n_iter),
                    lambda_reg=float(lambda_reg),
                    robust=bool(robust),
                    gradient_weighting=bool(gradient_weighting),
                    return_bias_field=True,
                    zero_background=bool(zero_background),
                )
                save_nifti(str(obf), bias.astype(np.float32), affine)
                logger.info(f"Bias field saved as {obf}")

            save_nifti(corrected_out_path, corrected_data, affine, hdr=img.header)
            logger.info(f"Corrected volume saved as {corrected_out_path}")

        return io_it



import numpy as np

from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import transform_streamlines
from dipy.utils.optpkg import optional_package

fury, have_fury, setup_module = optional_package(
    "fury", min_version="0.10.0", max_version="1.0.0"
)

if have_fury:
    import fury.io
    import fury.utils
    import vtk
    import vtk.util.numpy_support as ns

    DATATYPE_DICT = {
        np.dtype("int8"): vtk.VTK_CHAR,
        np.dtype("uint8"): vtk.VTK_UNSIGNED_CHAR,
        np.dtype("int16"): vtk.VTK_SHORT,
        np.dtype("uint16"): vtk.VTK_UNSIGNED_SHORT,
        np.dtype("int32"): vtk.VTK_INT,
        np.dtype("uint32"): vtk.VTK_UNSIGNED_INT,
        np.dtype("int64"): vtk.VTK_LONG_LONG,
        np.dtype("uint64"): vtk.VTK_UNSIGNED_LONG_LONG,
        np.dtype("float32"): vtk.VTK_FLOAT,
        np.dtype("float64"): vtk.VTK_DOUBLE,
    }


def load_polydata(file_name):
    """Load a vtk polydata to a supported format file.

    Supported file formats are OBJ, VTK, VTP, FIB, PLY, STL and XML

    Parameters
    ----------
    file_name : string or Path

    Returns
    -------
    output : vtkPolyData

    """
    return fury.io.load_polydata(str(file_name))


@warning_for_keywords()
def save_polydata(
    polydata, file_name, *, binary=False, color_array_name=None, legacy_vtk_format=False
):
    """Save a vtk polydata to a supported format file.

    Save formats can be VTK, VTP, FIB, PLY, STL and XML.
    Color array can be saved as well either by being already in the polydata
    or by passing it as an argument (color_array).

    Parameters
    ----------
    polydata : vtkPolyData
        The polydata object to be saved.
    file_name : string or Path
        The output filename (.vtk, .fib, .ply, .stl and .xml)
    binary : bool, optional
        Save the file in binary format.
    color_array_name : str, optional
        Name of the color array to save.
    legacy_vtk_format : bool, optional
        Use legacy VTK file format. Only applied when fury >= 2.0
        is installed.
    """
    # use kwargs for backward compatibility with fury < 2.0
    kwargs = {}
    if fury.__version__.split(".")[0] >= "2":
        kwargs.update({"legacy_vtk_format": legacy_vtk_format})

    fury.io.save_polydata(
        polydata=polydata,
        file_name=str(file_name),
        binary=binary,
        color_array_name=color_array_name,
        **kwargs,
    )


@warning_for_keywords()
def save_vtk_streamlines(streamlines, filename, *, to_lps=True, binary=False):
    """Save streamlines as vtk polydata to a supported format file.

    File formats can be OBJ, VTK, VTP, FIB, PLY, STL and XML

    Parameters
    ----------
    streamlines : list
        list of 2D arrays or ArraySequence
    filename : string or Path
        output filename (.obj, .vtk, .fib, .ply, .stl and .xml)
    to_lps : bool
        Default to True, will follow the vtk file convention for streamlines
        Will be supported by MITKDiffusion and MI-Brain
    binary : bool
        save the file as binary

    """
    if to_lps:
        # ras (mm) to lps (mm)
        to_lps = np.eye(4)
        to_lps[0, 0] = -1
        to_lps[1, 1] = -1
        streamlines = transform_streamlines(streamlines, to_lps)

    polydata, _ = fury.utils.lines_to_vtk_polydata(streamlines)
    save_polydata(polydata, file_name=filename, binary=binary)


@warning_for_keywords()
def load_vtk_streamlines(filename, *, to_lps=True):
    """Load streamlines from vtk polydata.

    Load formats can be VTK, FIB

    Parameters
    ----------
    filename : string or Path
        input filename (.vtk or .fib)
    to_lps : bool
        Default to True, will follow the vtk file convention for streamlines
        Will be supported by MITK-Diffusion and MI-Brain

    Returns
    -------
    output :  list
         list of 2D arrays

    """
    polydata = load_polydata(filename)
    lines = fury.utils.get_polydata_lines(polydata)
    if to_lps:
        to_lps = np.eye(4)
        to_lps[0, 0] = -1
        to_lps[1, 1] = -1
        return transform_streamlines(lines, to_lps)

    return lines


def get_polydata_triangles(polydata, dtype=None):
    """Get triangles from a vtkPolyData object.

    Parameters
    ----------
    polydata : vtkPolyData
        The polydata object from which to extract triangles.
    dtype : data-type, optional
        The desired data type for the output triangles. If None, the default
        data type will be used.
    Returns
    -------
    triangles : numpy.ndarray
        An array of shape (n_triangles, 3) containing the vertex indices
        of the triangles.

    """
    vtk_polys = ns.vtk_to_numpy(polydata.GetPolys().GetData())
    if len(vtk_polys) == 0:
        nbr_cells = polydata.GetNumberOfCells()
        triangles = []
        for i in range(nbr_cells):
            ids = polydata.GetCell(i).GetPointIds()
            for j in range(ids.GetNumberOfIds() - 2):
                triangles.append([ids.GetId(j), ids.GetId(j + 1), ids.GetId(j + 2)])
        triangles = np.array(triangles)
    else:
        if not (vtk_polys[::4] == 3).all():
            raise ValueError("Not all polygons are triangles")
        triangles = np.vstack([vtk_polys[1::4], vtk_polys[2::4], vtk_polys[3::4]]).T
    if dtype is not None:
        return triangles.astype(dtype)
    return triangles


def get_polydata_vertices(polydata, dtype=None):
    """Get vertices from a vtkPolyData object.

    Parameters
    ----------
    polydata : vtkPolyData
        The polydata object from which to extract vertices.
    dtype : data-type, optional
        The desired data type for the output vertices. If None, the default
        data type will be used.
    Returns
    -------
    vertices : numpy.ndarray
        An array of shape (n_vertices, 3) containing the vertex coordinates.

    """
    vertices = ns.vtk_to_numpy(polydata.GetPoints().GetData())
    if dtype is not None:
        return vertices.astype(dtype)
    return vertices


def convert_to_polydata(vertices, triangles, data_per_point=None):
    """Convert vertices and triangles to a vtkPolyData object.

    Parameters
    ----------
    vertices : numpy.ndarray
        An array of shape (n_vertices, 3) containing the vertex coordinates.
    triangles : numpy.ndarray
        An array of shape (n_triangles, 3) containing the vertex indices
        of the triangles.
    data_per_point : dict, optional
        A dictionary where keys are array names and values are numpy arrays
        of shape (n_vertices, ...) representing data associated with each
        vertex.

    Returns
    -------
    polydata : vtkPolyData
        The resulting vtkPolyData object.
    """
    vtk_points = vtk.vtkPoints()
    vtk_points.SetData(ns.numpy_to_vtk(vertices, deep=True))

    vtk_triangles = np.hstack(np.c_[np.ones(len(triangles)).astype(int) * 3, triangles])
    vtk_triangles = ns.numpy_to_vtkIdTypeArray(vtk_triangles, deep=True)
    vtk_cells = vtk.vtkCellArray()
    vtk_cells.SetCells(len(triangles), vtk_triangles)

    polydata = vtk.vtkPolyData()
    polydata.SetPoints(vtk_points)
    polydata.SetPolys(vtk_cells)

    if data_per_point is not None:
        for name, array in data_per_point.items():
            if len(array) != polydata.GetNumberOfPoints():
                raise ValueError("Array length does not match number of points.")
            vtk_array = _numpy_to_vtk_array(np.array(array), name=name)

            if "normal" in name.lower():
                polydata.GetPointData().SetNormals(vtk_array)
            else:
                polydata.GetPointData().AddArray(vtk_array)

    return polydata


def _numpy_to_vtk_array(array, name=None, dtype=None, deep=True):
    """Convert a numpy array to a vtk array.

    Parameters
    ----------
    array : numpy.ndarray
        The input numpy array.
    name : str, optional
        The name to assign to the vtk array.
    dtype : data-type, optional
        The desired data type for the vtk array. If None, the default data
        type of the numpy array will be used.
    deep : bool, optional
        If True, a deep copy of the data is made.

    Returns
    -------
    vtk_array : vtkDataArray
        The resulting vtk array.
    """
    if dtype is not None:
        vtk_dtype = DATATYPE_DICT[np.dtype(dtype)]
    else:
        vtk_dtype = DATATYPE_DICT[np.dtype(array.dtype)]
    vtk_array = ns.numpy_to_vtk(np.asarray(array), deep=True, array_type=vtk_dtype)
    if name is not None:
        vtk_array.SetName(name)
    return vtk_array
